import json
from typing import Optional, List, Any, Iterable, TYPE_CHECKING, Tuple, Union
from .base_endpoint import BaseEndpoint
from ..pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..util.api_call_decorator import api_call
from ..util.deprecation_decorator import deprecated_param
from ..util.text_enum import TextEnum

if TYPE_CHECKING:
    from boxsdk.object.folder import Folder
    from boxsdk.object.user import User
    from boxsdk.object.item import Item
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class SearchScope(TextEnum):
    """Enum of possible serach scopes."""
    USER = 'user_content'
    ENTERPRISE = 'enterprise_content'


class TrashContent(TextEnum):
    """Enum of trash content values."""
    NONE = 'non_trashed_only'
    ONLY = 'trashed_only'


class MetadataSearchFilter:
    """
    Helper class to encapsulate a single search filter. A search filter can only search against one template,
    but can filter on many fields.
    See :class:`MetadataSearchFilters`.
    """
    def __init__(self, template_key: str, scope: str):
        """
        :param template_key:
            The key of the template to search on
        :param scope:
            The scope of the template to search on
        """
        self._template_key = template_key
        self._scope = scope
        self._field_filters = {}

    def as_dict(self) -> dict:
        """
        Returns a `dict` representation of this object

        :return:
            The `dict` representation
        """
        return {
            'templateKey': self._template_key,
            'scope': self._scope,
            'filters': self._field_filters
        }

    def add_value_based_filter(self, field_key: str, value: str) -> None:
        """
        Add a value-based filter (used for token-based search on string fields, and exact match search on all other fields)

        :param field_key:
            The field key to filter on
        :param value:
            The value to use to filter
        """
        self._field_filters.update({field_key: value})

    def add_range_filter(
            self,
            field_key: str,
            gt_value: Union[str, int, float] = None,
            lt_value: Union[str, int, float] = None
    ) -> None:
        """
        Add a range filter (used for ranged searches on numbers and dates)

        :param field_key:
            The field key to filter on
        :param gt_value:
            The lower bound of the range filter (inclusive)
        :param lt_value:
            The upper bound of the range filter (inclusive)
        """
        range_part = {}
        if gt_value:
            range_part['gt'] = gt_value
        if lt_value:
            range_part['lt'] = lt_value
        if not range_part:
            raise ValueError('Should specify gt and/or lt')
        self._field_filters.update({field_key: range_part})


class MetadataSearchFilters:
    """
    Helper class to encapsulate a list of metadata search filter params (mdfilters API param)
    See https://developers.box.com/metadata-api/#search for more details
    """
    def __init__(self):
        self._filters = []

    def as_list(self) -> List[dict]:
        """
        Get a list of filters from this object to use as a parameter in the Search API

        :return:
            The list of filters
        """
        return [metadata_filter.as_dict() for metadata_filter in self._filters]

    def add_filter(self, metadata_filter: MetadataSearchFilter) -> None:
        """
        Add a filter to this object. Note that the API only supports one filter.

        :param metadata_filter:
            The filter to add
        """
        self._filters.append(metadata_filter)


class Search(BaseEndpoint):
    """Search Box for files and folders."""

    def get_url(self, *args: Any) -> str:
        """
        Gets the search endpoint URL.

        :return:
            The search endpoint URL.
        """
        return super().get_url('search', *args)

    @staticmethod
    def start_metadata_filters() -> MetadataSearchFilters:
        """
        Get a :class:`MetadataSearchFilters` that represents a set of metadata filters.

        :return:
            The new :class:`MetadataSearchFilters`
        """
        return MetadataSearchFilters()

    @staticmethod
    def make_single_metadata_filter(template_key: str, scope: str) -> MetadataSearchFilter:
        """
        Make a single :class:`MetadataSearchFilter` that represents a filter on a template. It must be
        added to a :class:`MetadataSearchFilters`.
        :param template_key:
            The key of the template to filter on
        :param scope:
            The scope of the template to filter on
        :return:
            The new :class:`MetadataSearchFilter`
        """
        return MetadataSearchFilter(template_key, scope)

    @api_call
    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches
    def query(
            self,
            query: str,
            limit: int = None,
            offset: int = 0,
            ancestor_folders: Iterable['Folder'] = None,
            file_extensions: Iterable[str] = None,
            metadata_filters: MetadataSearchFilters = None,
            result_type: str = None,
            content_types: Iterable[str] = None,
            scope: Optional[str] = None,
            created_at_range: Tuple[Optional[str], Optional[str]] = None,
            updated_at_range: Tuple[Optional[str], Optional[str]] = None,
            size_range: Tuple[Optional[int], Optional[int]] = None,
            owner_users: Iterable['User'] = None,
            trash_content: Optional[str] = None,
            fields: Iterable[str] = None,
            sort: Optional[str] = None,
            direction: Optional[str] = None,
            **kwargs: Any
    ) -> Iterable['Item']:
        """
        Search Box for items matching the given query.

        :param query:
            The string to search for.
        :param limit:
            The maximum number of items to return.
        :param offset:
            The search result at which to start the response.
        :param ancestor_folders:
            Folder ids to limit the search to.
        :param file_extensions:
            File extensions to limit the search to.
        :param metadata_filters:
            Filters used for metadata search
        :param result_type:
            Which type of result you want. Can be file or folder.
        :param content_types:
            Which content types to search. Valid types include name, description, file_content, comments, and tags.
        :param scope:
            The scope of content to search over
        :param created_at_range:
            A tuple of the form (lower_bound, upper_bound) for the creation datetime of items to search.
        :param updated_at_range:
            A tuple of the form (lower_bound, upper_bound) for the update datetime of items to search.
        :param size_range:
            A tuple of the form (lower_bound, upper_bound) for the size in bytes of items to search.
        :param owner_users:
            Owner users to filter content by; only content belonging to these users will be returned.
        :param trash_content:
            Whether to search trashed or non-trashed content.
        :param fields:
            Fields to include on the returned items.
        :param sort:
            What to sort the search results by. Currently `modified_at`
        :param direction:
            The direction to display the sorted search results. Can be set to `DESC` for descending or `ASC` for ascending.
        :return:
            The collection of items that match the search query.
        """
        url = self.get_url()
        additional_params = {'query': query}
        if ancestor_folders is not None:
            additional_params['ancestor_folder_ids'] = ','.join([folder.object_id for folder in ancestor_folders])
        if file_extensions is not None:
            additional_params['file_extensions'] = ','.join(file_extensions)
        if metadata_filters is not None:
            additional_params['mdfilters'] = json.dumps(metadata_filters.as_list())
        if content_types is not None:
            additional_params['content_types'] = ','.join(content_types)
        if result_type is not None:
            additional_params['type'] = result_type
        if scope is not None:
            additional_params['scope'] = scope
        if created_at_range is not None:
            additional_params['created_at_range'] = f'{created_at_range[0] or ""},{created_at_range[1] or ""}'
        if updated_at_range is not None:
            additional_params['updated_at_range'] = f'{updated_at_range[0] or ""},{updated_at_range[1] or ""}'
        if size_range is not None:
            additional_params['size_range'] = f'{size_range[0] or ""},{size_range[1] or ""}'
        if owner_users is not None:
            additional_params['owner_user_ids'] = ','.join([user.object_id for user in owner_users])
        if trash_content is not None:
            additional_params['trash_content'] = trash_content
        if sort is not None:
            additional_params['sort'] = sort
        if direction is not None:
            additional_params['direction'] = direction

        additional_params.update(kwargs)

        return LimitOffsetBasedObjectCollection(
            self._session,
            url,
            limit=limit,
            offset=offset,
            fields=fields,
            additional_params=additional_params,
            return_full_pages=False,
        )

    @deprecated_param(name="use_index", position=5, message="Parameter will be ignored. See docs for details.")
    @api_call
    def metadata_query(
            self,
            from_template: str,
            ancestor_folder_id: str,
            query: Optional[str] = None,
            query_params: Optional[dict] = None,
            use_index: Optional[str] = None,
            order_by: List[dict] = None,
            marker: Optional[str] = None,
            limit: int = None,
            fields: Iterable[Optional[str]] = None
    ) -> 'BoxObjectCollection':
        # pylint:disable=unused-argument
        """Query Box items by their metadata.

        :param from_template:
            The template used in the query. Must be in the form scope.templateKey.
        :param ancestor_folder_id:
            The folder_id to which to restrain the query
        :param query:
            The logical expression of the query
        :param query_params:
            Required if query present. The arguments for the query.
        :param use_index is deprecated
        :param order_by:
            The field_key(s) to order on and the corresponding direction(s)
        :param marker:
            The marker to use for requesting the next page
        :param limit:
            Max results to return for a single request (0-100 inclusive)
        :param fields:
            List of fields to request
        :returns:
            An iterator of the item search results
        """
        url = super().get_url('metadata_queries/execute_read')
        data = {
            'from': from_template,
            'ancestor_folder_id': ancestor_folder_id
        }
        if query is not None:
            data['query'] = query
        if query_params is not None:
            data['query_params'] = query_params
        if order_by is not None:
            data['order_by'] = order_by

        return MarkerBasedObjectCollection(
            session=self._session,
            url=url,
            limit=limit,
            marker=marker,
            fields=fields,
            additional_params=data,
            return_full_pages=False,
            use_post=True
        )

    @api_call
    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches
    def query_with_shared_links(
            self,
            query: str,
            limit: int = None,
            offset: int = 0,
            ancestor_folders: Iterable['Folder'] = None,
            file_extensions: Iterable[str] = None,
            metadata_filters: MetadataSearchFilters = None,
            result_type: str = None,
            content_types: Iterable[str] = None,
            scope: Optional[str] = None,
            created_at_range: Tuple[Optional[str], Optional[str]] = None,
            updated_at_range: Tuple[Optional[str], Optional[str]] = None,
            size_range: Tuple[Optional[int], Optional[int]] = None,
            owner_users: Iterable['User'] = None,
            trash_content: Optional[str] = None,
            fields: Iterable[str] = None,
            sort: Optional[str] = None,
            direction: Optional[str] = None,
            **kwargs: Any
    ) -> Iterable['Item']:
        """
        Search Box for items matching the given query. May also include items that are only accessible via recently used shared links.

        :param query:
            The string to search for.
        :param limit:
            The maximum number of items to return.
        :param offset:
            The search result at which to start the response.
        :param ancestor_folders:
            Folder ids to limit the search to.
        :param file_extensions:
            File extensions to limit the search to.
        :param metadata_filters:
            Filters used for metadata search
        :param result_type:
            Which type of result you want. Can be file or folder.
        :param content_types:
            Which content types to search. Valid types include name, description, file_content, comments, and tags.
        :param scope:
            The scope of content to search over
        :param created_at_range:
            A tuple of the form (lower_bound, upper_bound) for the creation datetime of items to search.
        :param updated_at_range:
            A tuple of the form (lower_bound, upper_bound) for the update datetime of items to search.
        :param size_range:
            A tuple of the form (lower_bound, upper_bound) for the size in bytes of items to search.
        :param owner_users:
            Owner users to filter content by; only content belonging to these users will be returned.
        :param trash_content:
            Whether to search trashed or non-trashed content.
        :param fields:
            Fields to include on the returned items.
        :param sort:
            What to sort the search results by. Currently `modified_at`
        :param direction:
            The direction to display the sorted search results. Can be set to `DESC` for descending or `ASC` for ascending.
        :return:
            The collection of items that match the search query.
        """
        url = self.get_url()
        additional_params = {'query': query, 'include_recent_shared_links': True}
        if ancestor_folders is not None:
            additional_params['ancestor_folder_ids'] = ','.join([folder.object_id for folder in ancestor_folders])
        if file_extensions is not None:
            additional_params['file_extensions'] = ','.join(file_extensions)
        if metadata_filters is not None:
            additional_params['mdfilters'] = json.dumps(metadata_filters.as_list())
        if content_types is not None:
            additional_params['content_types'] = ','.join(content_types)
        if result_type is not None:
            additional_params['type'] = result_type
        if scope is not None:
            additional_params['scope'] = scope
        if created_at_range is not None:
            additional_params['created_at_range'] = f'{created_at_range[0] or ""},{created_at_range[1] or ""}'
        if updated_at_range is not None:
            additional_params['updated_at_range'] = f'{updated_at_range[0] or ""},{updated_at_range[1] or ""}'
        if size_range is not None:
            additional_params['size_range'] = f'{size_range[0] or ""},{size_range[1] or ""}'
        if owner_users is not None:
            additional_params['owner_user_ids'] = ','.join([user.object_id for user in owner_users])
        if trash_content is not None:
            additional_params['trash_content'] = trash_content
        if sort is not None:
            additional_params['sort'] = sort
        if direction is not None:
            additional_params['direction'] = direction

        additional_params.update(kwargs)

        return LimitOffsetBasedObjectCollection(
            self._session,
            url,
            limit=limit,
            offset=offset,
            fields=fields,
            additional_params=additional_params,
            return_full_pages=False,
        )



# pylint: disable=too-many-lines
import datetime
import json
from io import BytesIO
from unittest.mock import Mock, ANY

import pytest
import pytz

from boxsdk.auth.oauth2 import OAuth2, TokenScope
from boxsdk.client import Client, DeveloperTokenClient, DevelopmentClient, LoggingClient
from boxsdk.config import API
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.object.collaboration import Collaboration
from boxsdk.object.collaboration_allowlist import CollaborationAllowlist
from boxsdk.object.email_alias import EmailAlias
from boxsdk.object.collection import Collection
from boxsdk.object.comment import Comment
from boxsdk.object.device_pinner import DevicePinner
from boxsdk.object.enterprise import Enterprise
from boxsdk.object.events import Events
from boxsdk.object.file_request import FileRequest
from boxsdk.object.folder import Folder
from boxsdk.object.file import File
from boxsdk.object.file_version import FileVersion
from boxsdk.object.group import Group
from boxsdk.object.invite import Invite
from boxsdk.object.storage_policy import StoragePolicy
from boxsdk.object.storage_policy_assignment import StoragePolicyAssignment
from boxsdk.object.terms_of_service import TermsOfService
from boxsdk.object.user import User
from boxsdk.object.upload_session import UploadSession
from boxsdk.object.trash import Trash
from boxsdk.object.group_membership import GroupMembership
from boxsdk.object.metadata_template import MetadataTemplate, MetadataField, MetadataFieldType
from boxsdk.object.retention_policy import RetentionPolicy
from boxsdk.object.retention_policy_assignment import RetentionPolicyAssignment
from boxsdk.object.file_version_retention import FileVersionRetention
from boxsdk.object.legal_hold import LegalHold
from boxsdk.object.legal_hold_policy import LegalHoldPolicy
from boxsdk.object.legal_hold_policy_assignment import LegalHoldPolicyAssignment
from boxsdk.object.metadata_cascade_policy import MetadataCascadePolicy
from boxsdk.object.sign_request import SignRequest
from boxsdk.object.sign_template import SignTemplate
from boxsdk.object.task import Task
from boxsdk.object.task_assignment import TaskAssignment
from boxsdk.object.webhook import Webhook
from boxsdk.object.web_link import WebLink
from boxsdk.pagination.marker_based_object_collection import MarkerBasedObjectCollection
from boxsdk.util.datetime_formatter import normalize_date_to_rfc3339_format


@pytest.fixture
def developer_token_input(monkeypatch):
    monkeypatch.setattr('builtins.input', lambda prompt: 'developer_token')


@pytest.fixture(params=[Client, DeveloperTokenClient, DevelopmentClient, LoggingClient])
def mock_client(mock_box_session, developer_token_input, request):
    # pylint:disable=redefined-outer-name, unused-argument
    mock_oauth = Mock(OAuth2)
    client = request.param(mock_oauth)
    # pylint:disable=protected-access
    client._session = mock_box_session
    return client


@pytest.fixture(scope='module')
def user_id_1():
    return 1


@pytest.fixture(scope='module')
def user_id_2():
    return 1023


@pytest.fixture(scope='module')
def file_id():
    return 100


@pytest.fixture(scope='module')
def folder_id():
    return '1022'


@pytest.fixture()
def test_folder(mock_box_session, mock_object_id):
    return Folder(mock_box_session, mock_object_id)


@pytest.fixture()
def test_webhook(mock_box_session, mock_object_id):
    return Webhook(mock_box_session, mock_object_id)


@pytest.fixture(scope='function')
def mock_file_response(mock_object_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'file', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture(scope='function')
def mock_folder_response(mock_object_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'folder', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture(scope='function')
def mock_content_response(make_mock_box_request):
    mock_box_response, mock_network_response = make_mock_box_request(content=b'Contents of a text file.')
    mock_network_response.response_as_stream = raw = Mock()
    raw.stream.return_value = (bytes((b,)) for b in mock_box_response.content)
    return mock_box_response


@pytest.fixture(scope='module')
def marker_id():
    return 'marker_1'


@pytest.fixture(scope='module')
def users_response(user_id_1, user_id_2):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'user', 'id': user_id_1},
            {'type': 'user', 'id': user_id_2}
        ],
        'limit': 100,
        'offset': 0,
        'total_count': 2
    }
    return mock_network_response


@pytest.fixture(scope='module')
def user_response(user_id_1):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {'type': 'user', 'id': user_id_1}
    return mock_network_response


@pytest.fixture(scope='module')
def group_id_1():
    return 101


@pytest.fixture(scope='module')
def group_id_2():
    return 202


@pytest.fixture()
def mock_user(mock_box_session, mock_user_id):
    user = User(mock_box_session, mock_user_id)
    return user


@pytest.fixture()
def mock_user_list(mock_box_session):
    first_user = User(mock_box_session, '33333')
    second_user = User(mock_box_session, '44444')
    user_list = [first_user, second_user]
    return user_list


@pytest.fixture()
def mock_file(mock_box_session):
    test_file = File(mock_box_session, '11111')
    return test_file


@pytest.fixture()
def mock_retention_policy(mock_box_session):
    retention_policy = RetentionPolicy(mock_box_session, '22222')
    return retention_policy


@pytest.fixture(scope='module')
def groups_response(group_id_1, group_id_2):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'group', 'id': group_id_1, 'name': str(group_id_1)},
            {'type': 'group', 'id': group_id_2, 'name': str(group_id_2)},
        ],
        'limit': 100,
        'offset': 0,
        'total_count': 2
    }
    return mock_network_response


@pytest.fixture(scope='module')
def legal_hold_policy_id_1():
    return 101


@pytest.fixture(scope='module')
def legal_hold_policy_id_2():
    return 202


@pytest.fixture(scope='module')
def legal_hold_policies_response(legal_hold_policy_id_1, legal_hold_policy_id_2):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'legal_hold_policy', 'id': legal_hold_policy_id_1, 'name': 'Test Policy 1'},
            {'type': 'legal_hold_policy', 'id': legal_hold_policy_id_2, 'name': 'Test Policy 2'},
        ],
        'limit': 5,
    }
    return mock_network_response


@pytest.fixture(scope='module')
def create_policy_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'type': 'legal_hold_policy',
        'id': 1234,
        'policy_name': 'Test Policy'
    }
    return mock_network_response


@pytest.fixture(scope='module')
def create_group_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'type': 'group',
        'id': 1234,
        'name': 'test_group_name',
    }
    return mock_network_response


@pytest.fixture(scope='module')
def tos_id_1():
    return 101


@pytest.fixture(scope='module')
def tos_id_2():
    return 202


@pytest.fixture(scope='module')
def terms_of_services_response(tos_id_1, tos_id_2):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'terms_of_service', 'id': tos_id_1},
            {'type': 'terms_of_service', 'id': tos_id_2},
        ],
        'total_count': 2,
    }
    return mock_network_response


@pytest.fixture(scope='module')
def create_user_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'type': 'user',
        'id': 1234,
        'name': 'Ned Stark',
    }
    return mock_network_response


@pytest.fixture(scope='module')
def create_invite_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'type': 'invite',
        'id': 1234,
    }
    return mock_network_response


@pytest.fixture(params=('file', 'folder'))
def test_item_and_response(mock_file, test_folder, mock_file_response, mock_folder_response, request):
    if request.param == 'file':
        return mock_file, mock_file_response
    return test_folder, mock_folder_response


@pytest.fixture()
def create_webhook_response(test_item_and_response, test_webhook):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'type': test_webhook.object_type,
        'id': test_webhook.object_id,
        'target': {
            'type': test_item.object_type,
            'id': test_item.object_id,
        },
        'created_at': '2016-05-09T17:41:27-07:00',
        'address': 'https://test.com',
        'triggers': [
            'FILE.UPLOADED',
            'FOLDER.CREATED',
        ],
    }
    return mock_network_response


@pytest.fixture(scope='module', params=('file', 'folder'))
def shared_item_response(request):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'type': request.param,
        'id': 1234,
        'name': 'shared_item',
    }
    return mock_network_response


@pytest.fixture(scope='module')
def search_response(file_id, folder_id):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'file', 'id': file_id},
            {'type': 'folder', 'id': folder_id}
        ],
        'limit': 100,
        'offset': 0,
        'total_count': 2
    }
    return mock_network_response


@pytest.fixture(scope='module')
def recent_items_response(file_id):
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'recent_item', 'item': {'type': 'file', 'id': file_id}}
        ],
        'next_marker': None,
        'limit': 100,
    }
    return mock_network_response


@pytest.fixture(scope='module')
def device_pin_id_1():
    return 101


@pytest.fixture(scope='module')
def device_pin_id_2():
    return 202


@pytest.fixture(scope='module')
def device_pins_response(device_pin_id_1, device_pin_id_2):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'device_pinner', 'id': device_pin_id_1},
            {'type': 'device_pinner', 'id': device_pin_id_2},
        ],
        'limit': 2,
    }
    return mock_network_response


@pytest.mark.parametrize('test_class, factory_method_name', [
    (EmailAlias, 'email_alias'),
    (Enterprise, 'enterprise'),
    (Folder, 'folder'),
    (File, 'file'),
    (FileVersion, 'file_version'),
    (Invite, 'invite'),
    (User, 'user'),
    (Group, 'group'),
    (GroupMembership, 'group_membership'),
    (Enterprise, 'enterprise'),
    (Webhook, 'webhook'),
    (MetadataCascadePolicy, 'metadata_cascade_policy'),
    (UploadSession, 'upload_session'),
    (StoragePolicy, 'storage_policy'),
    (StoragePolicyAssignment, 'storage_policy_assignment'),
])
def test_factory_returns_the_correct_object(mock_client, test_class, factory_method_name):
    """ Tests the various id-only factory methods in the Client class """
    # pylint:disable=redefined-outer-name
    fake_id = 'fake_id'

    factory_method = getattr(mock_client, factory_method_name)

    obj = factory_method(fake_id)

    assert isinstance(obj, test_class)
    assert obj.object_id == fake_id


def test_root_folder(mock_client):
    folder = mock_client.root_folder()
    assert isinstance(folder, Folder)
    assert folder.object_id == '0'


@pytest.fixture(scope='module', params=(None, 'user1'))
def users_filter_term(request):
    return request.param


@pytest.fixture(scope='module', params=(0, 10))
def users_offset(request):
    return request.param


@pytest.fixture(scope='module', params=(0, 10))
def users_limit(request):
    return request.param


@pytest.fixture(scope='module', params=(None, 'all', 'external', 'managed'))
def users_type(request):
    return request.param


def test_users_return_the_correct_user_objects(
        mock_client,
        mock_box_session,
        users_response,
        user_id_1,
        user_id_2,
        users_filter_term,
        users_type,
        users_offset,
        users_limit,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value = users_response
    users = mock_client.users(limit=users_limit, offset=users_offset, filter_term=users_filter_term, user_type=users_type)
    expected_params = {'offset': users_offset}
    if users_limit is not None:
        expected_params['limit'] = users_limit
    if users_filter_term is not None:
        expected_params['filter_term'] = users_filter_term
    if users_type is not None:
        expected_params['user_type'] = users_type
    assert users.next().object_id == user_id_1
    assert users.next().object_id == user_id_2
    mock_box_session.get.assert_called_once_with(f'{API.BASE_API_URL}/users', params=expected_params)


def test_users_return_the_correct_user_objects_marker(
        mock_client,
        mock_box_session,
        user_id_1,
        user_id_2,
        users_filter_term,
        users_type,
        marker_id,
        users_limit,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value.json.return_value = {
        'entries': [
            {'type': 'user', 'id': user_id_1},
            {'type': 'user', 'id': user_id_2}
        ],
        'next_marker': 'JGWUGKNJFSAH123NDSA',
        'total_count': 2
    }
    users = mock_client.users(limit=users_limit, marker=marker_id, use_marker=True, filter_term=users_filter_term, user_type=users_type)
    expected_params = {'marker': marker_id, 'usemarker': True}
    if users_limit is not None:
        expected_params['limit'] = users_limit
    if users_filter_term is not None:
        expected_params['filter_term'] = users_filter_term
    if users_type is not None:
        expected_params['user_type'] = users_type
    assert users.next().object_id == user_id_1
    assert users.next().object_id == user_id_2
    mock_box_session.get.assert_called_once_with(f'{API.BASE_API_URL}/users', params=expected_params)


def test_users_returns_correct_with_default_values(
        mock_client,
        mock_box_session,
):
    expected_url = f'{API.BASE_API_URL}/users'
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'offset': 0,
        'total_count': 1,
        'entries': [
            {
                'type': 'user',
                'id': '12345',
            }
        ]
    }
    users = mock_client.users()
    user = users.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'offset': 0})
    assert isinstance(user, User)
    assert user.type == 'user'
    assert user.id == '12345'


def test_search_instantiates_search_and_calls_search(
        mock_client,
        mock_box_session,
        search_response,
        file_id,
        folder_id,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value = search_response
    search_term = 'lolcatz'
    search_result = mock_client.search().query(
        search_term,
        10,
        0,
        ancestor_folders=[Folder(mock_box_session, folder_id)],
        file_extensions=['.jpg'],
    )
    assert search_result.next().object_id == file_id
    assert search_result.next().object_id == folder_id


def test_events_returns_event_object(mock_client):
    # pylint:disable=redefined-outer-name
    assert isinstance(mock_client.events(), Events)


def test_collaboration_allowlist_initializer(mock_client):
    collaboration_allowlist = mock_client.collaboration_allowlist()
    assert isinstance(collaboration_allowlist, CollaborationAllowlist)


def test_get_groups_return_the_correct_group_objects(
        mock_client,
        mock_box_session,
        groups_response,
        group_id_1,
        group_id_2,
):
    # pylint:disable=redefined-outer-name
    group_name = 'Employees'
    expected_url = f'{API.BASE_API_URL}/groups'
    mock_box_session.get.return_value = groups_response
    groups = mock_client.get_groups(group_name)
    for group, expected_id in zip(groups, [group_id_1, group_id_2]):
        assert group.object_id == expected_id
        assert group.name == str(expected_id)
        # pylint:disable=protected-access
        assert group._session == mock_box_session
    mock_box_session.get.assert_called_once_with(expected_url, params={'offset': None, 'filter_term': group_name})


def test_create_group_returns_the_correct_group_object(mock_client, mock_box_session, create_group_response):
    # pylint:disable=redefined-outer-name
    expected_url = f'{API.BASE_API_URL}/groups'
    test_group_name = 'test_group_name'
    value = json.dumps({
        'name': test_group_name,
        'provenance': 'Example',
        'external_sync_identifier': 'Example-User',
        'description': 'Description of group',
        'invitability_level': 'admins_and_members',
        'member_viewability_level': 'admins_only',
    })
    mock_box_session.post.return_value = create_group_response
    new_group = mock_client.create_group(
        name=test_group_name,
        provenance='Example',
        external_sync_identifier='Example-User',
        description='Description of group',
        invitability_level='admins_and_members',
        member_viewability_level='admins_only',
        fields=['name,description'],
    )

    mock_box_session.post.assert_called_once_with(expected_url, data=value, params={'fields': 'name,description'})
    assert isinstance(new_group, Group)
    assert new_group.object_id == 1234
    assert new_group.name == test_group_name


@pytest.mark.parametrize('description', ('My test policy',))
@pytest.mark.parametrize('filter_started_at', ('2016-01-01T00:00:00+00:00', datetime.datetime(2016, 1, 1, tzinfo=pytz.UTC)))
@pytest.mark.parametrize('filter_ended_at', ('2020-01-01T00:00:00+00:00', datetime.datetime(2020, 1, 1, tzinfo=pytz.UTC)))
@pytest.mark.parametrize('is_ongoing', ('True',))
def test_create_legal_hold_policy_returns_the_correct_policy_object(
        mock_client,
        mock_box_session,
        create_policy_response,
        description,
        filter_started_at,
        filter_ended_at,
        is_ongoing
):
    # pylint:disable=redefined-outer-name
    params = {
        'description': description,
        'filter_started_at': filter_started_at,
        'filter_ended_at': filter_ended_at,
        'is_ongoing': is_ongoing
    }

    test_policy_name = 'Test Policy'
    create_policy_response.json.return_value.update(params)
    mock_box_session.post.return_value = create_policy_response

    new_policy = mock_client.create_legal_hold_policy(test_policy_name, **params)

    expected_url = f'{API.BASE_API_URL}/legal_hold_policies'
    expected_body = {
        'policy_name': test_policy_name,
        'description': description,
        'filter_started_at': normalize_date_to_rfc3339_format(filter_started_at),
        'filter_ended_at': normalize_date_to_rfc3339_format(filter_ended_at),
        'is_ongoing': is_ongoing
    }
    mock_box_session.post.assert_called_once_with(expected_url, data=ANY)
    assert dict(json.loads(mock_box_session.post.call_args[1]['data'])) == expected_body
    assert isinstance(new_policy, LegalHoldPolicy)
    assert new_policy.policy_name == test_policy_name
    for param, expected_value in params.items():
        assert new_policy[param] == expected_value


def test_get_legal_hold_policies_return_the_correct_policy_objects(
        mock_client,
        mock_box_session,
        legal_hold_policies_response,
        legal_hold_policy_id_1,
        legal_hold_policy_id_2,
):
    # pylint:disable=redefined-outer-name
    policy_name = 'Arbitration'
    expected_url = f'{API.BASE_API_URL}/legal_hold_policies'
    mock_box_session.get.return_value = legal_hold_policies_response
    policies = mock_client.get_legal_hold_policies(policy_name)
    for policy, expected_id in zip(policies, [legal_hold_policy_id_1, legal_hold_policy_id_2]):
        assert policy.object_id == expected_id
        # pylint:disable=protected-access
        assert policy._session == mock_box_session
    mock_box_session.get.assert_called_once_with(expected_url, params={'policy_name': policy_name})


def test_trash_initializer(mock_client):
    trash = mock_client.trash()
    assert isinstance(trash, Trash)


def test_get_recent_items_returns_the_correct_items(mock_client, mock_box_session, recent_items_response, file_id):
    mock_box_session.get.return_value = recent_items_response
    recent_items = mock_client.get_recent_items()
    assert isinstance(recent_items, MarkerBasedObjectCollection)
    recent_item = recent_items.next()
    assert recent_item.item.object_id == file_id
    next_pointer = recent_items.next_pointer()
    assert next_pointer is None


def test_get_recent_items_sends_get_with_correct_params(mock_client, mock_box_session, recent_items_response, marker_id):
    limit = 50
    marker = marker_id
    fields = ['modified_at', 'name']
    expected_params = {
        'limit': limit,
        'marker': marker_id,
        'fields': ','.join(fields),
    }
    mock_box_session.get.return_value = recent_items_response
    object_collection = mock_client.get_recent_items(limit=limit, marker=marker, fields=fields)
    object_collection.next()
    mock_box_session.get.assert_called_once_with(f'{API.BASE_API_URL}/recent_items', params=expected_params)


@pytest.mark.parametrize('password', (None, 'p4ssw0rd'))
def test_get_shared_item_returns_the_correct_item(mock_client, mock_box_session, shared_item_response, password):
    # pylint:disable=redefined-outer-name
    shared_link = 'https://cloud.box.com/s/661wcw2iz6q5r7v5xxkm'
    mock_box_session.request.return_value = shared_item_response
    item = mock_client.get_shared_item(shared_link, password)
    assert item.type == shared_item_response.json()['type']
    mock_box_session.request.assert_called_once_with(
        'GET',
        f'{API.BASE_API_URL}/shared_items',
        headers={
            'BoxApi': f'shared_link={shared_link}{f"&shared_link_password={password}" if password is not None else ""}'
        },
    )


@pytest.mark.parametrize('test_method', [
    'get',
    'post',
    'put',
    'delete',
    'options',
])
def test_make_request_passes_request_on_to_session(mock_client, mock_box_session, test_method):
    # pylint:disable=redefined-outer-name
    mock_client.make_request(test_method, 'url')
    assert mock_box_session.request.call_args[0] == (test_method, 'url')


def test_create_app_user_returns_the_correct_user_object(mock_client, mock_box_session, create_user_response):
    # pylint:disable=redefined-outer-name
    test_user_name = 'Ned Stark'
    value = json.dumps({'name': test_user_name, 'is_platform_access_only': True})
    mock_box_session.post.return_value = create_user_response
    new_user = mock_client.create_user(name=test_user_name)

    assert len(mock_box_session.post.call_args_list) == 1

    assert mock_box_session.post.call_args[0] == (f'{API.BASE_API_URL}/users',)
    assert mock_box_session.post.call_args[1] == {'data': value}
    assert isinstance(new_user, User)
    assert new_user.object_id == 1234
    assert new_user.name == test_user_name


def test_create_enterprise_user_returns_the_correct_user_object(mock_client, mock_box_session, create_user_response):
    # pylint:disable=redefined-outer-name
    test_user_name = 'Ned Stark'
    test_user_login = 'eddard@box.com'
    value = json.dumps({'name': test_user_name, 'login': test_user_login})
    mock_box_session.post.return_value = create_user_response
    new_user = mock_client.create_user(name=test_user_name, login=test_user_login)

    assert len(mock_box_session.post.call_args_list) == 1

    assert mock_box_session.post.call_args[0] == (f"{API.BASE_API_URL}/users",)
    assert mock_box_session.post.call_args[1] == {'data': value}
    assert isinstance(new_user, User)
    assert new_user.object_id == 1234
    assert new_user.name == test_user_name


def test_get_storage_policies(mock_client, mock_box_session):
    expected_url = mock_box_session.get_url('storage_policies')
    mock_policy = {
        'type': 'storage_policy',
        'id': '12345',
        'name': 'Test Storage Policy'
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_policy]
    }
    policies = mock_client.get_storage_policies()
    policy = policies.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(policy, StoragePolicy)
    assert policy.type == 'storage_policy'
    assert policy.id == '12345'
    assert policy.name == 'Test Storage Policy'


def test_create_terms_of_service(mock_client, mock_box_session):
    # pylint:disable=redefined-outer-name
    expected_url = f'{API.BASE_API_URL}/terms_of_services'
    test_text = 'This is a test text'
    test_tos_type = 'external'
    test_status = 'enabled'
    value = json.dumps({
        'status': 'enabled',
        'tos_type': 'external',
        'text': 'This is a test text',
    })
    mock_box_session.post.return_value.json.return_value = {
        'type': 'terms_of_service',
        'id': '12345',
        'status': test_status,
        'tos_type': test_tos_type,
        'text': test_text,
    }
    new_terms_of_service = mock_client.create_terms_of_service('enabled', 'external', 'This is a test text')
    mock_box_session.post.assert_called_once_with(expected_url, data=value)
    assert isinstance(new_terms_of_service, TermsOfService)
    assert new_terms_of_service.type == 'terms_of_service'
    assert new_terms_of_service.id == '12345'
    assert new_terms_of_service.status == test_status
    assert new_terms_of_service.tos_type == test_tos_type
    assert new_terms_of_service.text == test_text


def test_get_all_terms_of_services(mock_client, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/terms_of_services'
    tos_body = {
        'type': 'terms_of_service',
        'id': '12345',
        'status': 'enabled',
        'tos_type': 'external',
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 1,
        'entries': [tos_body],
    }
    services = mock_client.get_terms_of_services(tos_type='external')
    service = services.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'tos_type': 'external'})
    assert isinstance(service, TermsOfService)
    assert service.type == 'terms_of_service'
    assert service.id == '12345'
    assert service.status == 'enabled'
    assert service.tos_type == 'external'


def test_create_webhook_returns_the_correct_policy_object(
        test_item_and_response,
        test_webhook,
        mock_client,
        mock_box_session,
        create_webhook_response,
):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/webhooks'
    expected_body = {
        'target': {
            'type': test_item.object_type,
            'id': test_item.object_id,
        },
        'triggers': ['FILE.UPLOADED', 'FOLDER.CREATED'],
        'address': 'https://test.com',
    }
    value = json.dumps(expected_body)
    mock_box_session.post.return_value = create_webhook_response
    new_webhook = mock_client.create_webhook(test_item, ['FILE.UPLOADED', 'FOLDER.CREATED'], 'https://test.com')
    mock_box_session.post.assert_called_once_with(
        expected_url,
        data=value,
    )
    assert isinstance(new_webhook, Webhook)
    assert new_webhook.id == test_webhook.object_id
    assert new_webhook.type == test_webhook.object_type
    assert new_webhook.target['type'] == test_item.object_type
    assert new_webhook.target['id'] == test_item.object_id
    assert new_webhook.triggers == ['FILE.UPLOADED', 'FOLDER.CREATED']
    assert new_webhook.address == 'https://test.com'


def test_get_webhooks(mock_client, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/webhooks'
    webhook_body = {
        'type': 'webhook',
        'id': '12345',
        'target': {
            'type': 'folder',
            'id': '11111',
        },
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [webhook_body],
    }
    webhooks = mock_client.get_webhooks()
    webhook = webhooks.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(webhook, Webhook)
    assert webhook.object_id == webhook_body['id']
    assert webhook.object_type == webhook_body['type']
    assert webhook.target['id'] == webhook_body['target']['id']
    assert webhook.target['type'] == webhook_body['target']['type']


def test_create_retention_policy(mock_client, mock_box_session, mock_user_list):
    policy_name = 'Test Retention Policy'
    policy_type = 'finite'
    disposition_action = 'remove_retention'
    policy_description = 'Test Retention Policy'
    expected_url = f'{API.BASE_API_URL}/retention_policies'
    expected_data = {
        'policy_name': policy_name,
        'disposition_action': disposition_action,
        'policy_type': 'finite',
        'retention_length': 5,
        'can_owner_extend_retention': True,
        'are_owners_notified': False,
        'custom_notification_recipients': [
            {
                'type': mock_user_list[0].object_type,
                'id': mock_user_list[0].object_id,
            },
            {
                'type': mock_user_list[1].object_type,
                'id': mock_user_list[1].object_id,
            },
        ],
        'retention_type': 'modifiable',
        'description': policy_description,
    }
    mock_policy = {
        'type': 'retention_policy',
        'id': '1234',
        'policy_name': policy_name,
        'policy_type': policy_type,
        'retention_length': 5,
        'disposition_action': disposition_action,
        'can_owner_extend_retention': False,
        'are_owners_notified': False,
        'custom_notification_recipients': [
            {
                'type': mock_user_list[0].object_type,
                'id': mock_user_list[0].object_id,
            },
            {
                'type': mock_user_list[1].object_type,
                'id': mock_user_list[1].object_id,
            },
        ],
        'retention_type': 'modifiable',
        'description': policy_description,
    }
    mock_box_session.post.return_value.json.return_value = mock_policy
    policy = mock_client.create_retention_policy(
        policy_name=policy_name,
        disposition_action=disposition_action,
        retention_length=5,
        can_owner_extend_retention=True,
        are_owners_notified=False,
        custom_notification_recipients=mock_user_list,
        retention_type='modifiable',
        description=policy_description,
    )
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert policy.object_id == mock_policy['id']
    assert policy.object_type == mock_policy['type']
    assert policy.policy_name == mock_policy['policy_name']
    assert policy.disposition_action == mock_policy['disposition_action']
    assert policy.can_owner_extend_retention == mock_policy['can_owner_extend_retention']
    assert policy.are_owners_notified == mock_policy['are_owners_notified']
    assert policy.retention_type == mock_policy['retention_type']
    assert policy.description == mock_policy['description']
    assert isinstance(policy, RetentionPolicy)


def test_create_infinte_retention_policy(mock_client, mock_box_session):
    policy_name = 'Test Retention Policy'
    policy_type = 'indefinite'
    disposition_action = 'remove_retention'
    policy_description = 'Test Retention Policy'
    expected_url = f'{API.BASE_API_URL}/retention_policies'
    expected_data = {
        'policy_name': policy_name,
        'disposition_action': disposition_action,
        'policy_type': policy_type,
        'can_owner_extend_retention': False,
        'are_owners_notified': False,
        'retention_type': 'non_modifiable',
        'description': policy_description,
    }
    mock_policy = {
        'type': 'retention_policy',
        'id': '1234',
        'policy_name': policy_name,
        'policy_type': policy_type,
        'disposition_action': disposition_action,
        'can_owner_extend_retention': False,
        'are_owners_notified': False,
        'retention_type': 'non_modifiable',
        'description': policy_description,
    }
    mock_box_session.post.return_value.json.return_value = mock_policy
    policy = mock_client.create_retention_policy(
        policy_name=policy_name,
        disposition_action=disposition_action,
        retention_length=float('inf'),
        can_owner_extend_retention=False,
        are_owners_notified=False,
        retention_type='non_modifiable',
        description=policy_description,
    )
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert policy.object_id == mock_policy['id']
    assert policy.object_type == mock_policy['type']
    assert policy.policy_name == mock_policy['policy_name']
    assert policy.disposition_action == mock_policy['disposition_action']
    assert policy.can_owner_extend_retention == mock_policy['can_owner_extend_retention']
    assert policy.are_owners_notified == mock_policy['are_owners_notified']
    assert policy.retention_type == mock_policy['retention_type']
    assert policy.description == mock_policy['description']
    assert isinstance(policy, RetentionPolicy)


def test_get_retention_policies(mock_client, mock_box_session, mock_user, mock_user_id):
    expected_url = f'{API.BASE_API_URL}/retention_policies'
    mock_policy = {
        'type': 'retention_policy',
        'id': '12345',
        'name': 'Test Retention Policy',
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_policy],
        'next_marker': 'testMarker',
    }
    policies = mock_client.get_retention_policies(policy_name='Test Name', policy_type='finite', user=mock_user)
    policy = policies.next()
    params = {
        'policy_name': 'Test Name',
        'policy_type': 'finite',
        'created_by_user_id': mock_user_id,
    }
    mock_box_session.get.assert_called_once_with(expected_url, params=params)
    assert isinstance(policy, RetentionPolicy)
    assert policy.id == mock_policy['id']
    assert policy.name == mock_policy['name']


def test_get_file_version_retentions(mock_client, mock_box_session, mock_file, mock_retention_policy):
    expected_url = f'{API.BASE_API_URL}/file_version_retentions'
    mock_retention = {
        'type': 'file_version_retention',
        'id': '12345',
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_retention],
        'next_marker': 'testMarker',
    }
    retentions = mock_client.get_file_version_retentions(
        target_file=mock_file,
        policy=mock_retention_policy,
        disposition_action='remove_retention',
        disposition_before='2014-09-15T13:15:35-07:00',
        disposition_after='2014-09-20T13:15:35-07:00',
    )
    retention = retentions.next()
    params = {
        'file_id': '11111',
        'policy_id': '22222',
        'disposition_action': 'remove_retention',
        'disposition_before': '2014-09-15T13:15:35-07:00',
        'disposition_after': '2014-09-20T13:15:35-07:00',
    }
    mock_box_session.get.assert_called_once_with(expected_url, params=params)
    assert isinstance(retention, FileVersionRetention)
    assert retention.id == mock_retention['id']
    assert retention.type == mock_retention['type']


def test_get_pending_collaborations(mock_client, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    expected_url = f'{API.BASE_API_URL}/collaborations'
    mock_collaboration = {
        'type': 'collaboration',
        'id': '12345',
        'created_by': {
            'type': 'user',
            'id': '33333',
        },
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 1,
        'limit': 2,
        'offset': 0,
        'entries': [mock_collaboration],
    }
    pending_collaborations = mock_client.get_pending_collaborations(limit=2)
    pending_collaboration = pending_collaborations.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'limit': 2, 'status': 'pending', 'offset': None})
    assert isinstance(pending_collaboration, Collaboration)
    assert pending_collaboration.id == mock_collaboration['id']
    assert pending_collaboration.type == mock_collaboration['type']
    assert pending_collaboration['created_by']['type'] == 'user'
    assert pending_collaboration['created_by']['id'] == '33333'


@pytest.fixture
def check_downscope_token_request(
        mock_client,
        mock_box_session,
        mock_object_id,
        make_mock_box_request,
):
    def do_check(scopes, item_class, additional_data, shared_link, expected_data):
        temp_downscoped_token = 'temp_downscoped_token'
        temp_expires_in = 1234
        mock_box_response, _ = make_mock_box_request(
            response={'access_token': temp_downscoped_token, 'expires_in': temp_expires_in},
        )
        mock_box_session.post.return_value = mock_box_response

        item = item_class(mock_box_session, mock_object_id) if item_class else None

        downscoped_token_response = mock_client.downscope_token(scopes, item, additional_data, shared_link)

        assert downscoped_token_response.access_token == temp_downscoped_token
        assert downscoped_token_response.expires_in == temp_expires_in

        if item:
            expected_data['resource'] = item.get_url()
        mock_box_session.post.assert_called_once_with(f'{API.OAUTH2_API_URL}/token', data=expected_data)

    return do_check


@pytest.mark.parametrize(
    'item_class,scopes,expected_scopes',
    [
        (File, [TokenScope.ITEM_READWRITE], 'item_readwrite'),
        (Folder, [TokenScope.ITEM_PREVIEW, TokenScope.ITEM_SHARE], 'item_preview item_share'),
        (File, [TokenScope.ITEM_READ, TokenScope.ITEM_SHARE, TokenScope.ITEM_DELETE], 'item_read item_share item_delete'),
        (None, [TokenScope.ITEM_DOWNLOAD], 'item_download'),
    ],
)
def test_downscope_token_sends_downscope_request(
        mock_client,
        check_downscope_token_request,
        item_class,
        scopes,
        expected_scopes,
):
    mock_client.auth.access_token = 'existing_access_token'
    expected_data = {
        'subject_token': 'existing_access_token',
        'subject_token_type': 'urn:ietf:params:oauth:token-type:access_token',
        'scope': expected_scopes,
        'grant_type': 'urn:ietf:params:oauth:grant-type:token-exchange',
    }
    check_downscope_token_request(scopes, item_class, None, None, expected_data)


def test_downscope_token_sends_downscope_request_with_shared_link(
        mock_client,
        check_downscope_token_request,
):
    shared_link = 'https://cloud.box.com/s/foo'
    expected_data = {
        'subject_token': mock_client.auth.access_token,
        'subject_token_type': 'urn:ietf:params:oauth:token-type:access_token',
        'scope': 'item_readwrite',
        'grant_type': 'urn:ietf:params:oauth:grant-type:token-exchange',
        'box_shared_link': shared_link
    }
    check_downscope_token_request([TokenScope.ITEM_READWRITE], None, None, shared_link, expected_data)


def test_downscope_token_sends_downscope_request_with_additional_data(
        mock_client,
        check_downscope_token_request,
):
    additional_data = {'grant_type': 'new_grant_type', 'extra_data_key': 'extra_data_value'}
    expected_data = {
        'subject_token': mock_client.auth.access_token,
        'subject_token_type': 'urn:ietf:params:oauth:token-type:access_token',
        'scope': 'item_readwrite',
        'grant_type': 'new_grant_type',
        'extra_data_key': 'extra_data_value',
    }
    check_downscope_token_request([TokenScope.ITEM_READWRITE], File, additional_data, None, expected_data)


def test_downscope_token_sends_downscope_request_when_no_initial_token(
        mock_client,
        check_downscope_token_request,
):
    mock_client.auth.access_token = None
    mock_client.auth.refresh.return_value = 'new_access_token'

    expected_data = {
        'subject_token': 'new_access_token',
        'subject_token_type': 'urn:ietf:params:oauth:token-type:access_token',
        'scope': 'item_readwrite',
        'grant_type': 'urn:ietf:params:oauth:grant-type:token-exchange',
    }
    check_downscope_token_request([TokenScope.ITEM_READWRITE], File, None, None, expected_data)
    mock_client.auth.refresh.assert_called_once_with(None)


def test_device_pins_for_enterprise(mock_client, mock_box_session, device_pins_response, device_pin_id_1, device_pin_id_2):
    # pylint:disable=redefined-outer-name
    enterprise_id = '11111'
    expected_url = f'{API.BASE_API_URL}/enterprises/{enterprise_id}/device_pinners'
    mock_box_session.get.return_value = device_pins_response
    enterprise = mock_client.enterprise(enterprise_id)
    pins = mock_client.device_pinners(enterprise, direction='ASC')
    for pin, expected_id in zip(pins, [device_pin_id_1, device_pin_id_2]):
        assert pin.object_id == expected_id
        # pylint:disable=protected-access
        assert pin._session == mock_box_session
    mock_box_session.get.assert_called_once_with(expected_url, params={'direction': 'ASC'})


def test_metadata_template_initializer(mock_client, mock_box_session):
    template = mock_client.metadata_template('enterprise', 'VendorContract')
    assert isinstance(template, MetadataTemplate)
    # pylint:disable=protected-access
    assert template._session == mock_box_session
    assert template.object_id is None
    assert template.scope == 'enterprise'
    assert template.template_key == 'VendorContract'


def test_metadata_template_by_id(mock_client, mock_box_session):
    template_id = 'sdkjfhgsdg-nb34745bndfg-qw4hbsajdg'

    template = mock_client.metadata_template_by_id(template_id)

    assert isinstance(template, MetadataTemplate)
    # pylint:disable=protected-access
    assert template._session == mock_box_session
    assert template.object_id == template_id
    assert template.scope is None
    assert template.template_key is None


def test_get_metadata_templates(mock_client, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_templates/enterprise'
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 1,
        'entries': [
            {
                'type': 'metadata_template',
                'scope': 'enterprise_33333',
                'displayName': 'Vendor Contract',
                'templateKey': 'vendorContract',
                'fields': [
                    {
                        'type': 'string',
                        'displayName': 'Name',
                        'key': 'name',
                    },
                ],
            },
        ],
        'next_marker': None,
        'previous_marker': None,
    }

    templates = mock_client.get_metadata_templates()
    template = templates.next()

    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(template, MetadataTemplate)
    assert template.object_id is None
    assert template.displayName == 'Vendor Contract'
    fields = template.fields
    assert len(fields) == 1
    field = fields[0]
    assert isinstance(field, dict)
    assert field['type'] == 'string'
    assert field['key'] == 'name'


def test_create_metadata_template(mock_client, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_templates/schema'
    name = 'Vendor Contract'
    key = 'vContract'
    field1 = MetadataField(MetadataFieldType.DATE, 'Birthday', 'bday')
    field2 = MetadataField(MetadataFieldType.ENUM, 'State', options=['CA', 'TX', 'NY'])
    expected_body = {
        'scope': 'enterprise',
        'displayName': 'Vendor Contract',
        'hidden': True,
        'fields': [
            {
                'type': 'date',
                'displayName': 'Birthday',
                'key': 'bday',
            },
            {
                'type': 'enum',
                'displayName': 'State',
                'options': [
                    {'key': 'CA'},
                    {'key': 'TX'},
                    {'key': 'NY'},
                ],
            },
        ],
        'copyInstanceOnItemCopy': False,
        'templateKey': 'vContract',
    }

    response = {
        'type': 'metadata_template',
    }
    response.update(expected_body)
    mock_box_session.post.return_value.json.return_value = response

    template = mock_client.create_metadata_template(name, [field1, field2], key, hidden=True)

    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert isinstance(template, MetadataTemplate)
    assert template.object_id is None
    assert template.displayName == 'Vendor Contract'
    fields = template.fields
    assert len(fields) == 2
    field = fields[0]
    assert isinstance(field, dict)
    assert field['type'] == 'date'
    assert field['key'] == 'bday'


def test_get_current_enterprise(mock_client, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/users/me'
    expected_params = {
        'fields': 'enterprise'
    }
    enterprise_id = '44444'
    enterprise_name = 'Acme, Inc.'
    user_json = {
        'type': 'user',
        'id': '33333',
        'enterprise': {
            'type': 'enterprise',
            'id': enterprise_id,
            'name': enterprise_name,
        },
    }
    mock_box_session.get.return_value.json.return_value = user_json

    enterprise = mock_client.get_current_enterprise()

    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params, headers=None)
    assert isinstance(enterprise, Enterprise)
    assert enterprise.object_id == enterprise_id
    assert enterprise._session == mock_box_session  # pylint:disable=protected-access
    assert enterprise.name == enterprise_name
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params={'fields': 'enterprise'})


def test_comment(mock_client):
    # pylint:disable=redefined-outer-name
    comment_id = '12345'
    comment = mock_client.comment(comment_id)

    assert isinstance(comment, Comment)
    assert comment.object_id == comment_id


def test_collaboration(mock_client):
    # pylint:disable=redefined-outer-name
    collaboration_id = '12345'
    collaboration = mock_client.collaboration(collaboration_id)

    assert isinstance(collaboration, Collaboration)
    assert collaboration.object_id == collaboration_id


def test_legal_hold_policy(mock_client):
    # pylint:disable=redefined-outer-name
    policy_id = '12345'
    policy = mock_client.legal_hold_policy(policy_id)

    assert isinstance(policy, LegalHoldPolicy)
    assert policy.object_id == policy_id


def test_legal_hold_policy_assignment(mock_client):
    # pylint:disable=redefined-outer-name
    assignment_id = '12345'
    assignment = mock_client.legal_hold_policy_assignment(assignment_id)

    assert isinstance(assignment, LegalHoldPolicyAssignment)
    assert assignment.object_id == assignment_id


def test_legal_hold(mock_client):
    # pylint:disable=redefined-outer-name
    hold_id = '12345'
    legal_hold = mock_client.legal_hold(hold_id)

    assert isinstance(legal_hold, LegalHold)
    assert legal_hold.object_id == hold_id


def test_collection(mock_client):
    # pylint:disable=redefined-outer-name
    collection_id = '12345'
    collection = mock_client.collection(collection_id)

    assert isinstance(collection, Collection)
    assert collection.object_id == collection_id


def test_collections(mock_client, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    expected_url = f'{API.BASE_API_URL}/collections'
    mock_collection = {
        'type': 'collection',
        'id': '12345',
        'created_by': {
            'type': 'user',
            'id': '33333',
        },
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 1,
        'limit': 2,
        'offset': 0,
        'entries': [mock_collection],
    }
    collections = mock_client.collections(limit=2)
    collection = collections.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'limit': 2, 'offset': 0})
    assert isinstance(collection, Collection)
    assert collection.id == mock_collection['id']
    assert collection.type == mock_collection['type']
    assert collection['created_by']['type'] == 'user'
    assert collection['created_by']['id'] == '33333'


def test_task(mock_client):
    # pylint:disable=redefined-outer-name
    task_id = '12345'
    task = mock_client.task(task_id)

    assert isinstance(task, Task)
    assert task.object_id == task_id


def test_task_assignment(mock_client):
    # pylint:disable=redefined-outer-name
    assignment_id = '12345'
    assignment = mock_client.task_assignment(assignment_id)

    assert isinstance(assignment, TaskAssignment)
    assert assignment.object_id == assignment_id


def test_retention_policy(mock_client):
    # pylint:disable=redefined-outer-name
    policy_id = '12345'
    policy = mock_client.retention_policy(policy_id)

    assert isinstance(policy, RetentionPolicy)
    assert policy.object_id == policy_id


def test_retention_policy_assignment(mock_client):
    # pylint:disable=redefined-outer-name
    assignment_id = '12345'
    assignment = mock_client.retention_policy_assignment(assignment_id)

    assert isinstance(assignment, RetentionPolicyAssignment)
    assert assignment.object_id == assignment_id


def test_file_version_retention(mock_client):
    # pylint:disable=redefined-outer-name
    retention_id = '12345'
    file_version_retention = mock_client.file_version_retention(retention_id)

    assert isinstance(file_version_retention, FileVersionRetention)
    assert file_version_retention.object_id == retention_id


def test_web_link(mock_client):
    # pylint:disable=redefined-outer-name
    web_link_id = '12345'
    web_link = mock_client.web_link(web_link_id)

    assert isinstance(web_link, WebLink)
    assert web_link.object_id == web_link_id


def test_device_pinner(mock_client):
    # pylint:disable=redefined-outer-name
    pin_id = '12345'
    pin = mock_client.device_pinner(pin_id)

    assert isinstance(pin, DevicePinner)
    assert pin.object_id == pin_id


def test_download_zip(mock_client, mock_box_session, mock_content_response):
    expected_create_url = f'{API.BASE_API_URL}/zip_downloads'
    name = 'test'
    file_item = mock_client.file('466239504569')
    folder_item = mock_client.folder('466239504580')
    items = [file_item, folder_item]
    mock_writeable_stream = BytesIO()
    expected_create_body = {
        'download_file_name': name,
        'items': [
            {
                'type': 'file',
                'id': '466239504569'
            },
            {
                'type': 'folder',
                'id': '466239504580'
            }
        ]
    }
    status_response_mock = Mock()
    status_response_mock.json.return_value = {
        'total_file_count': 20,
        'downloaded_file_count': 10,
        'skipped_file_count': 10,
        'skipped_folder_count': 10,
        'state': 'succeeded'
    }
    mock_box_session.post.return_value.json.return_value = {
        'download_url': 'https://dl.boxcloud.com/2.0/zip_downloads/124hfiowk3fa8kmrwh/content',
        'status_url': 'https://api.box.com/2.0/zip_downloads/124hfiowk3fa8kmrwh/status',
        'expires_at': '2018-04-25T11:00:18-07:00',
        'name_conflicts': [
            [
                {
                    'id': '100',
                    'type': 'file',
                    'original_name': 'salary.pdf',
                    'download_name': 'aqc823.pdf'
                },
                {
                    'id': '200',
                    'type': 'file',
                    'original_name': 'salary.pdf',
                    'download_name': 'aci23s.pdf'
                }
            ]
        ]
    }

    mock_box_session.get.side_effect = [mock_content_response, status_response_mock]

    status_returned = mock_client.download_zip(name, items, mock_writeable_stream)
    mock_box_session.post.assert_called_once_with(expected_create_url,
                                                  data=json.dumps(expected_create_body),
                                                  skip_retry_codes={202})
    mock_box_session.get.assert_any_call('https://dl.boxcloud.com/2.0/zip_downloads/124hfiowk3fa8kmrwh/content',
                                         expect_json_response=False, stream=True)
    mock_box_session.get.assert_called_with('https://api.box.com/2.0/zip_downloads/124hfiowk3fa8kmrwh/status')
    mock_writeable_stream.seek(0)
    assert mock_writeable_stream.read() == mock_content_response.content
    assert status_returned['total_file_count'] == 20
    assert status_returned['name_conflicts'][0][0]['id'] == '100'


@pytest.fixture(scope='module')
def mock_sign_request_response():
    # pylint:disable=redefined-outer-name
    mock_sign_request = {
        'id': '42',
        'type': 'sign-request',
        'are_reminders_enabled': 'true',
        'are_text_signatures_enabled': 'true',
        'auto_expire_at': '2021-04-26T08:12:13.982Z',
        'days_valid': '2',
        'email_message': 'Hello! Please sign the document below',
        'email_subject': 'Sign Request from Acme',
        'external_id': '123',
        'is_document_preparation_needed': 'true',
        'redirect_url': 'https://www.box.com/accepted',
        'declined_redirect_url': 'https://www.box.com/declined',
        'parent_folder': {
            'id': '12345',
            'type': 'folder',
            'etag': '1',
            'name': 'Contracts',
            'sequence_id': '3'
        },
        'prefill_tags': [
            {
                'document_tag_id': '1234',
                'text_value': 'text',
                'checkbox_value': 'true',
                'date_value': '2021-04-26T08:12:13.982Z'
            }
        ],
        'prepare_url': 'https://prepareurl.com',
        'sign_files': {
            'files': [
                {
                    'id': '12345',
                    'etag': '1',
                    'type': 'file',
                    'sequence_id': '3',
                    'name': 'Contract.pdf',
                    'sha1': '85136C79CBF9FE36BB9D05D0639C70C265C18D37',
                    'file_version': {
                        'id': '12345',
                        'type': 'file_version',
                        'sha1': '134b65991ed521fcfe4724b7d814ab8ded5185dc'
                    }
                }
            ],
            'is_ready_for_download': 'true'
        },
        'signers': [
            {
                'email': 'example@gmail.com',
                'role': 'signer',
                'is_in_person': 'true',
                'order': '2',
                'embed_url_external_user_id': '1234',
                'has_viewed_document': 'true',
                'signer_decision': {
                    'type': 'signed',
                    'finalized_at': '2021-04-26T08:12:13.982Z'
                },
                'inputs': [
                    {
                        'document_tag_id': '1234',
                        'text_value': 'text',
                        'checkbox_value': 'true',
                        'date_value': '2021-04-26T08:12:13.982Z',
                        'type': 'text',
                        'page_index': '4'
                    }
                ],
                'embed_url': 'https://example.com'
            }
        ],
        'signing_log': {
            'id': '12345',
            'type': 'file',
            'etag': '1',
            'file_version': {
                'id': '12345',
                'type': 'file_version',
                'sha1': '134b65991ed521fcfe4724b7d814ab8ded5185dc'
            },
            'name': 'Contract.pdf',
            'sequence_id': '3',
            'sha1': '85136C79CBF9FE36BB9D05D0639C70C265C18D37'
        },
        'source_files': [
            {
                'id': '12345',
                'etag': '1',
                'type': 'file',
                'sequence_id': '3',
                'name': 'Contract.pdf',
                'sha1': '85136C79CBF9FE36BB9D05D0639C70C265C18D37',
                'file_version': {
                    'id': '12345',
                    'type': 'file_version',
                    'sha1': '134b65991ed521fcfe4724b7d814ab8ded5185dc'
                }
            }
        ],
        'status': 'converting',
        'template_id': '123075213-af2c8822-3ef2-4952-8557-52d69c2fe9cb'
    }
    return mock_sign_request


@pytest.fixture(scope='module')
def mock_sign_template_response():
    mock_sign_template = {
        "id": "93153068-5420-467b-b8ef-8e54bfb7be42",
        "type": "sign-template",
        "name": "important-file.pdf",
        "email_message": "Please sign this document.\n\nKind regards",
        "email_subject": "Box User (boxuser@box.com) has requested your signature on a document",
        "parent_folder": {
            "id": "123456789",
            "etag": "0",
            "type": "folder",
            "sequence_id": "0",
            "name": "My Sign Requests"
        },
        "auto_expire_days": "null",
        "source_files": [
            {
                "id": "123456",
                "etag": "0",
                "type": "file",
                "sequence_id": "0",
                "sha1": "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                "file_version": {
                    "id": "123456",
                    "type": "file_version",
                    "sha1": "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                }
            }
        ],
        "are_email_settings_locked": "false",
        "are_fields_locked": "false",
        "are_files_locked": "false",
        "are_options_locked": "false",
        "are_recipients_locked": "false",
        "signers": [
            {
                "email": "",
                "label": "",
                "public_id": "AAQXQXJZ4",
                "role": "final_copy_reader",
                "is_in_person": "false",
                "order": 1,
                "inputs": []
            },
            {
                "email": "",
                "label": "",
                "public_id": "13XQXJZ4",
                "role": "signer",
                "is_in_person": "false",
                "order": 1,
                "inputs": [
                    {
                        "document_tag_id": None,
                        "id": "0260f921-3b52-477f-ae74-0b0b0b0b0b0b",
                        "type": "signature",
                        "text_value": None,
                        "is_required": True,
                        "coordinates": {
                            "x": 0.27038464059712,
                            "y": 0.10051756244533624
                        },
                        "dimensions": {
                            "width": 0.23570031566618235,
                            "height": 0.04781003891921971
                        },
                        "date_value": None,
                        "page_index": 0,
                        "checkbox_value": None,
                        "document_id": "2fdf9003-d798-40ee-be7f-0b0b0b0b0b0b",
                        "content_type": "signature",
                        "dropdown_choices": None,
                        "group_id": None,
                        "label": None
                    }
                ]
            }
        ],
        "ready_sign_link": None,
        "custom_branding": None,
        "days_valid": 0,
        "additional_info": {
            "non_editable": [],
            "required": {
                "signers": [
                    [
                        "email"
                    ],
                    [
                        "email"
                    ]
                ]
            }
        }
    }
    return mock_sign_template


def test_get_sign_requests(mock_client, mock_box_session, mock_sign_request_response):
    expected_url = f'{API.BASE_API_URL}/sign_requests'

    mock_sign_request = mock_sign_request_response
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 1,
        'limit': 100,
        'entries': [mock_sign_request],
        'next_marker': None,
        'previous_marker': None,
    }

    sign_requests = mock_client.get_sign_requests()
    sign_request = sign_requests.next()

    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(sign_request, SignRequest)
    assert sign_request.id == mock_sign_request['id']
    assert sign_request.status == mock_sign_request['status']


def test_create_sign_request(mock_client, mock_box_session, mock_sign_request_response):
    expected_url = f'{API.BASE_API_URL}/sign_requests'
    redirect_url = 'https://www.box.com/accepted'
    declined_redirect_url = 'https://www.box.com/declined'
    template_id = '123075213-af2c8822-3ef2-4952-8557-52d69c2fe9cb'
    source_file = {
        'id': '12345',
        'type': 'file'
    }
    source_file2 = {
        'id': '34567',
        'type': 'file'
    }
    files = [source_file, source_file2]

    signer = {
        'email': 'example@gmail.com'
    }
    signers = [signer]
    parent_folder_id = '12345'

    data = json.dumps({
        'source_files': [
            {
                'id': source_file['id'],
                'type': source_file['type']
            },
            {
                'id': source_file2['id'],
                'type': source_file2['type']
            }
        ],
        'signers': [
            {
                'email': signer['email']
            }
        ],
        'parent_folder':
        {
            'id': parent_folder_id,
            'type': 'folder'
        },
        'redirect_url': redirect_url,
        'declined_redirect_url': declined_redirect_url,
        'template_id': template_id
    })
    mock_box_session.post.return_value.json.return_value = mock_sign_request_response

    new_sign_request = mock_client.create_sign_request(
        files, signers, parent_folder_id,
        redirect_url=redirect_url, declined_redirect_url=declined_redirect_url, template_id=template_id)

    mock_box_session.post.assert_called_once_with(expected_url, data=data)
    assert isinstance(new_sign_request, SignRequest)
    assert new_sign_request['source_files'][0]['id'] == source_file['id']
    assert new_sign_request['signers'][0]['email'] == signer['email']
    assert new_sign_request['parent_folder']['id'] == parent_folder_id
    assert new_sign_request['redirect_url'] == redirect_url
    assert new_sign_request['declined_redirect_url'] == declined_redirect_url
    assert new_sign_request['template_id'] == template_id


def test_file_request(mock_client):
    # pylint:disable=redefined-outer-name
    file_request_id = '12345'
    file_request = mock_client.file_request(file_request_id)

    assert isinstance(file_request, FileRequest)
    assert file_request.object_id == file_request_id


def test_get_sign_templates_from_id(mock_client, mock_box_session, mock_sign_template_response):
    test_sign_template_id = '93153068-5420-467b-b8ef-8e54bfb7be42'
    expected_url = f'{API.BASE_API_URL}/sign_templates/{test_sign_template_id}'
    mock_box_session.get.return_value.json.return_value = mock_sign_template_response

    sign_template = mock_client.sign_template(test_sign_template_id).get()

    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)

    assert isinstance(sign_template, SignTemplate)
    assert sign_template.id == '93153068-5420-467b-b8ef-8e54bfb7be42'


def test_get_sign_template(mock_client, mock_box_session, mock_sign_template_response):
    test_sign_template_id = '93153068-5420-467b-b8ef-8e54bfb7be42'
    expected_url = f'{API.BASE_API_URL}/sign_templates/{test_sign_template_id}'
    mock_box_session.get.return_value.json.return_value = mock_sign_template_response

    sign_template = mock_client.get_sign_template(test_sign_template_id)

    mock_box_session.get.assert_called_once_with(expected_url)

    assert isinstance(sign_template, SignTemplate)
    assert sign_template.id == '93153068-5420-467b-b8ef-8e54bfb7be42'
    assert sign_template.name == 'important-file.pdf'
    assert sign_template.email_message == 'Please sign this document.\n\nKind regards'


def test_get_sign_templates(mock_client, mock_box_session, mock_sign_template_response):
    expected_url = f'{API.BASE_API_URL}/sign_templates'

    mock_sign_template = mock_sign_template_response
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 1,
        'limit': 100,
        'entries': [mock_sign_template],
        'next_marker': None,
        'previous_marker': None,
    }

    sign_templates = mock_client.get_sign_templates()

    sign_template = sign_templates.next()

    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(sign_template, SignTemplate)
    assert sign_template.id == '93153068-5420-467b-b8ef-8e54bfb7be42'
    assert sign_template.name == 'important-file.pdf'



import json
from datetime import datetime
from io import BytesIO
from unittest.mock import mock_open, patch, Mock

import pytest
import pytz
from pytest_lazyfixture import lazy_fixture

from boxsdk.config import API
from boxsdk.exception import BoxAPIException
from boxsdk.object.comment import Comment
from boxsdk.object.file import File
from boxsdk.object.file_version import FileVersion
from boxsdk.object.task import Task
from boxsdk.object.upload_session import UploadSession
from boxsdk.util.chunked_uploader import ChunkedUploader
from boxsdk.util.datetime_formatter import normalize_date_to_rfc3339_format
from boxsdk.util.default_arg_value import SDK_VALUE_NOT_SET


# pylint:disable=protected-access
# pylint:disable=too-many-lines
# pylint:disable=redefined-outer-name

@pytest.fixture()
def mock_accelerator_upload_url_for_update():
    return 'https://upload.box.com/api/2.0/files/fake_file_id/content?upload_session_id=123'


@pytest.fixture(scope='function')
def mock_accelerator_response_for_update(make_mock_box_request, mock_accelerator_upload_url_for_update):
    mock_response, _ = make_mock_box_request(
        response={
            'upload_url': mock_accelerator_upload_url_for_update,
            'download_url': None,
        }
    )
    return mock_response


def test_delete_file(test_file, mock_box_session, etag, if_match_header):
    test_file.delete(etag=etag)
    expected_url = test_file.get_url()
    mock_box_session.delete.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        params={},
        headers=if_match_header,
    )


def test_create_upload_session(test_file, mock_box_session):
    expected_url = f'{API.UPLOAD_URL}/files/{test_file.object_id}/upload_sessions'
    file_size = 197520
    part_size = 12345
    total_parts = 16
    num_parts_processed = 0
    upload_session_type = 'upload_session'
    upload_session_id = 'F971964745A5CD0C001BBE4E58196BFD'
    file_name = 'test_file.pdf'
    expected_data = {
        'file_id': test_file.object_id,
        'file_size': file_size,
        'file_name': file_name
    }
    mock_box_session.post.return_value.json.return_value = {
        'id': upload_session_id,
        'type': upload_session_type,
        'num_parts_processed': num_parts_processed,
        'total_parts': total_parts,
        'part_size': part_size,
    }
    upload_session = test_file.create_upload_session(file_size, file_name)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert isinstance(upload_session, UploadSession)
    assert upload_session._session == mock_box_session
    assert upload_session.part_size == part_size
    assert upload_session.total_parts == total_parts
    assert upload_session.num_parts_processed == num_parts_processed
    assert upload_session.type == upload_session_type
    assert upload_session.id == upload_session_id


def test_get_chunked_uploader(mock_box_session, mock_content_response, mock_file_path, test_file):
    expected_url = f'{API.UPLOAD_URL}/files/{test_file.object_id}/upload_sessions'
    mock_file_stream = BytesIO(mock_content_response.content)
    file_size = 197520
    part_size = 12345
    total_parts = 16
    num_parts_processed = 0
    upload_session_type = 'upload_session'
    upload_session_id = 'F971964745A5CD0C001BBE4E58196BFD'
    expected_data = {
        'file_id': test_file.object_id,
        'file_size': file_size,
    }
    mock_box_session.post.return_value.json.return_value = {
        'id': upload_session_id,
        'type': upload_session_type,
        'num_parts_processed': num_parts_processed,
        'total_parts': total_parts,
        'part_size': part_size,
    }
    with patch('os.stat') as stat:
        stat.return_value.st_size = file_size
        with patch('boxsdk.object.file.open', return_value=mock_file_stream):
            chunked_uploader = test_file.get_chunked_uploader(mock_file_path)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    upload_session = chunked_uploader._upload_session
    assert upload_session.part_size == part_size
    assert upload_session.total_parts == total_parts
    assert upload_session.num_parts_processed == num_parts_processed
    assert upload_session.type == upload_session_type
    assert upload_session.id == upload_session_id
    assert isinstance(chunked_uploader, ChunkedUploader)


@pytest.mark.parametrize(
    'due_at',
    [
        '2014-04-03T11:09:43+14:00',
        datetime(2014, 4, 3, 11, 9, 43, tzinfo=pytz.timezone('US/Alaska'))
    ]
)
def test_create_task(test_file, test_task, mock_box_session, due_at):
    # pylint:disable=redefined-outer-name
    expected_url = f"{API.BASE_API_URL}/tasks"
    action = 'review'
    message = 'Test Message'
    mock_box_session.post.return_value.json.return_value = {
        'type': test_task.object_type,
        'id': test_task.object_id,
        'due_at': '2014-04-03T11:09:43+14:00',
        'action': action,
        'message': message,
    }

    new_task = test_file.create_task(message=message, due_at=due_at)

    expected_body = {
        'item': {
            'type': 'file',
            'id': '42',
        },
        'action': action,
        'message': message,
        'due_at': '2014-04-03T11:09:43+14:00'
    }
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert isinstance(new_task, Task)
    assert new_task.object_type == test_task.object_type
    assert new_task.object_id == test_task.object_id
    assert new_task.action == action
    assert new_task.message == message
    assert new_task.due_at == '2014-04-03T11:09:43+14:00'


def test_create_task_with_review(test_file, test_task, mock_box_session):
    # pylint:disable=redefined-outer-name
    expected_url = f"{API.BASE_API_URL}/tasks"
    due_at = '2020-09-18T12:09:43+00:00'
    action = 'complete'
    message = 'Test Message'
    completion_rule = 'any_assignee'
    expected_body = {
        'item': {
            'type': 'file',
            'id': '42',
        },
        'action': action,
        'message': message,
        'due_at': due_at,
        'completion_rule': completion_rule,
    }
    mock_box_session.post.return_value.json.return_value = {
        'type': test_task.object_type,
        'id': test_task.object_id,
        'due_at': due_at,
        'action': action,
        'message': message,
        'completion_rule': completion_rule,
    }
    value = json.dumps(expected_body)
    new_task = test_file.create_task(
        message=message,
        due_at=due_at,
        action=action,
        completion_rule=completion_rule,
    )
    mock_box_session.post.assert_called_once_with(expected_url, data=value)
    assert isinstance(new_task, Task)
    assert new_task.object_type == test_task.object_type
    assert new_task.object_id == test_task.object_id
    assert new_task.action == action
    assert new_task.message == message
    assert new_task.due_at == due_at
    assert new_task.completion_rule == completion_rule


def test_get_tasks(test_file, mock_box_session):
    expected_url = test_file.get_url('tasks')
    task_body = {
        'type': 'task',
        'id': '12345',
        'item': {
            'type': 'file',
            'id': '33333',
        },
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [task_body],
    }
    tasks = test_file.get_tasks()
    task = tasks.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(task, Task)
    assert task.id == task_body['id']
    assert task.object_type == task_body['type']
    assert task.item['id'] == task_body['item']['id']


def test_get_download_url(test_file, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}/content'
    download_url = 'https://dl.boxcloud.com/sdjhfgksdjfgshdbg'
    mock_box_session.get.return_value.headers = {
        'location': download_url
    }
    url = test_file.get_download_url()
    mock_box_session.get.assert_called_once_with(
        expected_url,
        params=None,
        expect_json_response=False,
        allow_redirects=False
    )
    assert url == download_url


def test_get_download_url_file_version(test_file, test_file_version, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}/content'
    download_url = 'https://dl.boxcloud.com/sdjhfgksdjfgshdbg'
    mock_box_session.get.return_value.headers = {
        'location': download_url
    }
    url = test_file.get_download_url(file_version=test_file_version)
    mock_box_session.get.assert_called_once_with(
        expected_url,
        params={'version': test_file_version.object_id},
        expect_json_response=False,
        allow_redirects=False
    )
    assert url == download_url


@pytest.mark.parametrize('params,expected_query,expected_headers', [
    ({}, None, None),
    ({'byte_range': (100, 199)}, None, {'Range': 'bytes=100-199'}),
    ({'byte_range': (100,)}, None, {'Range': 'bytes=100-'}),
])
def test_download_to(test_file, mock_box_session, mock_content_response, params, expected_query, expected_headers):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}/content'
    mock_box_session.get.return_value = mock_content_response
    mock_writeable_stream = BytesIO()
    test_file.download_to(mock_writeable_stream, **params)
    mock_writeable_stream.seek(0)
    assert mock_writeable_stream.read() == mock_content_response.content
    mock_box_session.get.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        stream=True,
        params=expected_query,
        headers=expected_headers
    )


def test_download_to_file_version(test_file, test_file_version, mock_box_session, mock_content_response):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}/content'
    mock_box_session.get.return_value = mock_content_response
    mock_writeable_stream = BytesIO()
    test_file.download_to(mock_writeable_stream, file_version=test_file_version)
    mock_writeable_stream.seek(0)
    assert mock_writeable_stream.read() == mock_content_response.content
    mock_box_session.get.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        stream=True,
        headers=None,
        params={'version': test_file_version.object_id}
    )


@pytest.mark.parametrize('params,expected_query,expected_headers', [
    ({}, None, None),
    ({'byte_range': (100, 199)}, None, {'Range': 'bytes=100-199'}),
])
def test_get_content(test_file, mock_box_session, mock_content_response, params, expected_query, expected_headers):
    expected_url = test_file.get_url('content')
    mock_box_session.get.return_value = mock_content_response
    file_content = test_file.content(**params)
    assert file_content == mock_content_response.content
    mock_box_session.get.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        params=expected_query,
        headers=expected_headers
    )


def test_get_content_file_version(test_file, mock_box_session, mock_content_response, test_file_version):
    expected_url = test_file.get_url('content')
    mock_box_session.get.return_value = mock_content_response
    file_content = test_file.content(file_version=test_file_version)
    assert file_content == mock_content_response.content
    mock_box_session.get.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        params={'version': test_file_version.object_id},
        headers=None
    )


@pytest.mark.parametrize('content_modified_at', [
    '1970-01-01T11:11:11.500+14:00',
    datetime(1970, 1, 1, 11, 11, 11, microsecond=500, tzinfo=pytz.timezone('US/Alaska'))
])
@pytest.mark.parametrize('is_stream', (True, False))
def test_update_contents(
        test_file,
        mock_box_session,
        mock_content_response,
        mock_upload_response,
        mock_file_path,
        etag,
        sha1,
        upload_using_accelerator,
        mock_accelerator_response_for_update,
        mock_accelerator_upload_url_for_update,
        upload_using_accelerator_fails,
        if_match_sha1_header,
        is_stream,
        content_modified_at
):
    # pylint:disable=too-many-locals
    file_new_name = 'new_file_name'
    additional_attributes = {'attr': 123}
    expected_url = test_file.get_url('content').replace(API.BASE_API_URL, API.UPLOAD_URL)
    if upload_using_accelerator:
        if upload_using_accelerator_fails:
            mock_box_session.options.side_effect = BoxAPIException(400)
        else:
            mock_box_session.options.return_value = mock_accelerator_response_for_update
            expected_url = mock_accelerator_upload_url_for_update

    mock_box_session.post.return_value = mock_upload_response

    if is_stream:
        mock_file_stream = BytesIO(mock_content_response.content)
        new_file = test_file.update_contents_with_stream(
            mock_file_stream,
            etag=etag,
            upload_using_accelerator=upload_using_accelerator,
            file_name=file_new_name,
            content_modified_at=content_modified_at,
            additional_attributes=additional_attributes,
            sha1=sha1,
        )
    else:
        mock_file = mock_open(read_data=mock_content_response.content)
        mock_file_stream = mock_file.return_value
        with patch('boxsdk.object.file.open', mock_file, create=True):
            new_file = test_file.update_contents(
                mock_file_path,
                etag=etag,
                upload_using_accelerator=upload_using_accelerator,
                file_name=file_new_name,
                content_modified_at=content_modified_at,
                additional_attributes=additional_attributes,
                sha1=sha1,
            )

    mock_files = {'file': ('unused', mock_file_stream)}
    attributes = {
        'name': file_new_name,
        'content_modified_at': '1970-01-01T11:11:11+14:00',
    }
    # Using `update` to mirror the actual impl, since the attributes could otherwise come through in a different order
    # in Python 2 tests
    attributes.update(additional_attributes)
    data = {'attributes': json.dumps(attributes)}
    mock_box_session.post.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        files=mock_files,
        data=data,
        headers=if_match_sha1_header,
    )
    assert isinstance(new_file, File)
    assert new_file.object_id == test_file.object_id
    assert 'id' in new_file
    assert new_file['id'] == test_file.object_id
    assert not hasattr(new_file, 'entries')
    assert 'entries' not in new_file


@pytest.mark.parametrize('is_stream', (True, False))
def test_update_contents_combines_preflight_and_accelerator_calls_if_both_are_requested(
        test_file,
        mock_box_session,
        mock_file_path,
        mock_content_response,
        mock_accelerator_response_for_update,
        is_stream
):
    mock_box_session.options.return_value = mock_accelerator_response_for_update

    if is_stream:
        mock_file_stream = BytesIO(mock_content_response.content)
        test_file.update_contents_with_stream(
            mock_file_stream,
            preflight_check=True,
            upload_using_accelerator=True,
        )
    else:
        mock_file = mock_open(read_data=mock_content_response.content)
        with patch('boxsdk.object.file.open', mock_file, create=True):
            test_file.update_contents(
                mock_file_path,
                preflight_check=True,
                upload_using_accelerator=True,
            )

    mock_box_session.options.assert_called_once()


def test_update_contents_with_stream_does_preflight_check_if_specified(
        test_file,
        preflight_check,
        file_size,
        preflight_fails,
        mock_box_session,
):
    with patch.object(File, 'preflight_check', return_value=None):
        kwargs = {'file_stream': BytesIO(b'some bytes')}
        if preflight_check:
            kwargs['preflight_check'] = preflight_check
            kwargs['preflight_expected_size'] = file_size
        if preflight_fails:
            test_file.preflight_check.side_effect = BoxAPIException(400)
            with pytest.raises(BoxAPIException):
                test_file.update_contents_with_stream(**kwargs)
        else:
            test_file.update_contents_with_stream(**kwargs)

        if preflight_check:
            assert test_file.preflight_check.called_once_with(size=file_size)
            if preflight_fails:
                assert not mock_box_session.post.called
            else:
                assert mock_box_session.post.called
        else:
            assert not test_file.preflight_check.called


@patch('boxsdk.object.file.open', mock_open(read_data=b'some bytes'), create=True)
def test_update_contents_does_preflight_check_if_specified(
        test_file,
        mock_file_path,
        preflight_check,
        file_size,
        preflight_fails,
        mock_box_session,
):
    with patch.object(File, 'preflight_check', return_value=None):
        kwargs = {'file_path': mock_file_path}
        if preflight_check:
            kwargs['preflight_check'] = preflight_check
            kwargs['preflight_expected_size'] = file_size
        if preflight_fails:
            test_file.preflight_check.side_effect = BoxAPIException(400)
            with pytest.raises(BoxAPIException):
                test_file.update_contents(**kwargs)
        else:
            test_file.update_contents(**kwargs)

        if preflight_check:
            assert test_file.preflight_check.called_once_with(size=file_size)
            if preflight_fails:
                assert not mock_box_session.post.called
            else:
                assert mock_box_session.post.called
        else:
            assert not test_file.preflight_check.called


@pytest.mark.parametrize('params,expected_data', [
    ({}, {'is_download_prevented': False}),
    ({'prevent_download': False}, {'is_download_prevented': False}),
    ({'prevent_download': True}, {'is_download_prevented': True}),
    ({'expire_time': '2018-11-06T19:40:00-08:00'},
     {
        'is_download_prevented': False,
        'expires_at': '2018-11-06T19:40:00-08:00'
    }),
    ({'expire_time': datetime(2018, 11, 6, 19, 40, 00, tzinfo=pytz.timezone('US/Alaska'))},
     {
        'is_download_prevented': False,
        'expires_at': '2018-11-06T19:40:00+14:00'
    }),
])
def test_lock(test_file, mock_box_session, mock_file_response, params, expected_data):
    expected_url = test_file.get_url()
    expected_body = {
        'lock': {
            'type': 'lock'
        }
    }

    if 'is_download_prevented' in expected_data.keys():
        expected_body['lock']['is_download_prevented'] = expected_data['is_download_prevented']
    if 'expires_at' in expected_data.keys():
        expected_body['lock']['expires_at'] = expected_data['expires_at']

    mock_box_session.put.return_value = mock_file_response
    test_file.lock(**params)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_body),
        params=None,
        headers=None,
    )


def test_unlock(test_file, mock_box_session, mock_file_response):
    expected_url = test_file.get_url()
    mock_box_session.put.return_value = mock_file_response
    test_file.unlock()
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps({'lock': None}),
        params=None,
        headers=None,
    )


@pytest.mark.parametrize(
    'size, name, expected_data',
    [
        # Test case for specifying the name of the file for preflight
        (
            100,
            'foo.txt',
            json.dumps({'size': 100, 'name': 'foo.txt'}),
        ),

        # Test case for omitting the name of the file for preflight
        (
            200,
            None,
            json.dumps({'size': 200})
        ),
    ]
)
def test_preflight_check(
        test_file,
        mock_object_id,
        mock_box_session,
        mock_accelerator_response_for_update,
        mock_accelerator_upload_url_for_update,
        size,
        name,
        expected_data,
):
    mock_box_session.options.return_value = mock_accelerator_response_for_update
    kwargs = {'size': size}
    if name:
        kwargs['name'] = name

    accelerator_url = test_file.preflight_check(**kwargs)

    mock_box_session.options.assert_called_once_with(
        url=f'{API.BASE_API_URL}/files/{mock_object_id}/content',
        expect_json_response=True,
        data=expected_data,
    )
    assert accelerator_url == mock_accelerator_upload_url_for_update


def test_get_shared_link(
        test_file,
        mock_box_session,
        shared_link_access,
        shared_link_unshared_at,
        shared_link_password,
        shared_link_can_download,
        shared_link_can_preview,
        shared_link_can_edit,
        shared_link_vanity_name,
        test_url,
        etag,
        if_match_header,
):
    # pylint:disable=redefined-outer-name, protected-access
    expected_url = test_file.get_url()
    mock_box_session.put.return_value.json.return_value = {
        'type': test_file.object_type,
        'id': test_file.object_id,
        'shared_link': {
            'url': test_url,
        },
    }
    expected_data = {'shared_link': {}}
    if shared_link_access is not None:
        expected_data['shared_link']['access'] = shared_link_access
    if shared_link_unshared_at is not SDK_VALUE_NOT_SET:
        expected_data['shared_link']['unshared_at'] = normalize_date_to_rfc3339_format(shared_link_unshared_at)
    permissions = {}
    if shared_link_can_download is not None:
        permissions['can_download'] = shared_link_can_download
    if shared_link_can_preview is not None:
        permissions['can_preview'] = shared_link_can_preview
    if shared_link_can_edit is not None:
        permissions['can_edit'] = shared_link_can_edit
    if permissions:
        expected_data['shared_link']['permissions'] = permissions
    if shared_link_password is not None:
        expected_data['shared_link']['password'] = shared_link_password
    if shared_link_vanity_name is not None:
        expected_data['shared_link']['vanity_name'] = shared_link_vanity_name

    url = test_file.get_shared_link(
        etag=etag,
        access=shared_link_access,
        unshared_at=shared_link_unshared_at,
        password=shared_link_password,
        allow_download=shared_link_can_download,
        allow_preview=shared_link_can_preview,
        allow_edit=shared_link_can_edit,
        vanity_name=shared_link_vanity_name,
    )
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_data),
        headers=if_match_header,
        params=None,
    )
    assert url == test_url


def test_get_shared_link_download_url(
        test_file,
        mock_box_session,
        shared_link_access,
        shared_link_unshared_at,
        shared_link_password,
        shared_link_can_preview,
        shared_link_vanity_name,
        test_url,
        etag,
        if_match_header,
):
    # pylint:disable=redefined-outer-name, protected-access
    expected_url = test_file.get_url()
    mock_box_session.put.return_value.json.return_value = {
        'type': test_file.object_type,
        'id': test_file.object_id,
        'shared_link': {
            'url': None,
            'download_url': test_url,
        },
    }
    expected_data = {
        'shared_link': {},
    }
    if shared_link_access is not None:
        expected_data['shared_link']['access'] = shared_link_access
    if shared_link_unshared_at is not SDK_VALUE_NOT_SET:
        expected_data['shared_link']['unshared_at'] = normalize_date_to_rfc3339_format(shared_link_unshared_at)
    if shared_link_can_preview is not None:
        expected_data['shared_link']['permissions'] = permissions = {}
        permissions['can_preview'] = shared_link_can_preview
    if shared_link_password is not None:
        expected_data['shared_link']['password'] = shared_link_password
    if shared_link_vanity_name is not None:
        expected_data['shared_link']['vanity_name'] = shared_link_vanity_name

    url = test_file.get_shared_link_download_url(
        etag=etag,
        access=shared_link_access,
        unshared_at=shared_link_unshared_at,
        password=shared_link_password,
        allow_preview=shared_link_can_preview,
        vanity_name=shared_link_vanity_name,
    )
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_data),
        headers=if_match_header,
        params=None,
    )
    assert url == test_url


def test_get_comments(test_file, mock_box_session):
    expected_url = test_file.get_url('comments')
    mock_comment1 = {
        'type': 'comment',
        'id': '11111',
        'message': 'Foo'
    }
    mock_comment2 = {
        'type': 'comment',
        'id': '22222',
        'tagged_message': 'Well hello there, @[33333:friend]!'
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 2,
        'offset': 0,
        'limit': 100,
        'entries': [mock_comment1, mock_comment2]
    }
    comments = test_file.get_comments()
    comment1 = comments.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'offset': 0})
    assert comment1.object_id == mock_comment1['id']
    assert comment1.message == mock_comment1['message']

    comment2 = comments.next()
    assert comment2.object_id == mock_comment2['id']
    assert comment2.tagged_message == mock_comment2['tagged_message']


def test_add_comment(test_file, mock_box_session, comment_params):
    expected_url = 'https://api.box.com/2.0/comments'
    comment_id = '12345'
    (message_type, message) = comment_params
    expected_data = {
        message_type: message,
        'item': {
            'type': 'file',
            'id': test_file.object_id
        }
    }
    mock_box_session.post.return_value.json.return_value = {
        'type': 'comment',
        'id': comment_id,
        message_type: message
    }
    comment = test_file.add_comment(message)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert isinstance(comment, Comment)
    assert comment.object_id == comment_id


def test_get_previous_versions(test_file, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}/versions'
    mock_version1 = {
        'type': 'file_version',
        'id': '11111',
        'sha1': '4788db35f85f87acaaa5ba82cc99d72c9323281f',
    }
    mock_version2 = {
        'type': 'comment',
        'id': '22222',
        'sha1': '4788db35f85f87acaaa5ba82cc99d72c9323281f',
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 2,
        'offset': 0,
        'limit': 100,
        'entries': [mock_version1, mock_version2]
    }
    versions = test_file.get_previous_versions()
    version1 = versions.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'offset': None})
    assert version1.object_id == mock_version1['id']
    assert version1.sha1 == mock_version1['sha1']

    version2 = versions.next()
    assert version2.object_id == mock_version2['id']
    assert version2.sha1 == mock_version2['sha1']


def test_promote_version(test_file, test_file_version, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}/versions/current'
    sha1 = '12039d6dd9a7e6eefc78846802e'
    expected_body = {
        'type': 'file_version',
        'id': test_file_version.object_id,
    }
    mock_box_session.post.return_value.json.return_value = {
        'type': 'file_version',
        'id': '77777',
        'sha1': sha1,
    }
    new_version = test_file.promote_version(test_file_version)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert isinstance(new_version, FileVersion)
    assert new_version.object_id == '77777'
    assert new_version.sha1 == sha1


@pytest.mark.parametrize('params,expected_headers', [
    ({}, None),
    ({'etag': 'foobar'}, {'If-Match': 'foobar'}),
])
def test_delete_version(test_file, test_file_version, mock_box_session, params, expected_headers):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}/versions/{test_file_version.object_id}'
    mock_box_session.delete.return_value.ok = True
    is_success = test_file.delete_version(test_file_version, **params)
    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, headers=expected_headers)
    assert is_success is True


def test_get_embed_url(test_file, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}'
    expected_params = {
        'fields': 'expiring_embed_link'
    }
    embed_url = 'https://app.box.com/preview/mystuff'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'file',
        'id': test_file.object_id,
        'expiring_embed_link': {
            'url': embed_url,
        },
    }

    url = test_file.get_embed_url()
    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params)
    assert url == embed_url


@pytest.mark.parametrize('rep_hints,expected_headers', [
    (None, None),
    ('[pdf]', {'X-Rep-Hints': '[pdf]'}),
])
def test_get_representation_info(test_file, mock_box_session, rep_hints, expected_headers):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}'
    expected_params = {'fields': 'representations'}

    info_url = 'https://api.box.com/2.0/representations/pdf'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'file',
        'id': test_file.object_id,
        'representations': {
            'total_count': 1,
            'entries': [
                {
                    'representation': 'pdf',
                    'info': {
                        'url': info_url,
                    },
                },
            ],
        },
    }

    reps = test_file.get_representation_info(rep_hints=rep_hints)
    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params, headers=expected_headers)
    assert isinstance(reps, list)
    assert len(reps) == 1
    rep = reps[0]
    assert rep['representation'] == 'pdf'
    assert rep['info']['url'] == info_url


@pytest.mark.parametrize('extension,min_width,min_height,max_width,max_height,expected_params', [
    ('png', None, None, None, None, {}),
    ('png', None, None, None, None, {}),
    ('jpg', None, None, None, None, {}),
    ('png', 1, 2, None, None, {'min_width': 1, 'min_height': 2}),
    ('png', 1, 2, 3, 4, {'min_width': 1, 'min_height': 2, 'max_width': 3, 'max_height': 4}),
])
def test_get_thumbnail(
        test_file,
        mock_box_session,
        mock_content_response,
        extension,
        min_width,
        min_height,
        max_width,
        max_height,
        expected_params,
):
    expected_url = f'{API.BASE_API_URL}/files/{test_file.object_id}/thumbnail.{extension}'
    mock_box_session.get.return_value = mock_content_response

    thumb = test_file.get_thumbnail(
        extension=extension,
        min_width=min_width,
        min_height=min_height,
        max_width=max_width,
        max_height=max_height,
    )

    mock_box_session.get.assert_called_once_with(expected_url, expect_json_response=False, params=expected_params)
    assert thumb == mock_content_response.content


@pytest.mark.parametrize('dimensions,extension', [
    ('92x92', 'png'),
    ('92x92', 'jpg'),
])
def test_get_thumbnail_representation(
        test_file,
        mock_box_session,
        mock_content_response,
        dimensions,
        extension,
):
    representation_url = f'{API.BASE_API_URL}/files/{test_file.object_id}'
    content_url = 'https://dl.boxcloud.com/api/2.0/internal_files/123/versions/345/representations/jpg/content/'

    mock_representations_response = Mock()
    mock_representations_response.json.return_value = {
        'etag': '1',
        'id': test_file.object_id,
        'representations': {
            'entries': [
                {
                    'content': {
                        'url_template': content_url + '{+asset_path}'
                    },
                    'info': {
                        'url': 'https://api.box.com/2.0/internal_files/123/versions/345/representations/jpg'
                    },
                    'properties': {},
                    'representation': 'pdf',
                    'status': {
                        'state': 'success'
                    }
                }
            ]
        },
        'type': 'file'
    }

    mock_box_session.get.side_effect = [mock_representations_response, mock_content_response]

    thumb = test_file.get_thumbnail_representation(
        dimensions=dimensions,
        extension=extension,
    )

    mock_box_session.get.assert_any_call(representation_url, headers={'X-Rep-Hints': f'[{extension}?dimensions=92x92]'},
                                         params={'fields': 'representations'})
    mock_box_session.get.assert_any_call(content_url, expect_json_response=False)
    assert thumb == mock_content_response.content


def test_get_thumbnail_representation_not_found(
        test_file,
        mock_box_session,
        mock_content_response,
):
    representation_url = f'{API.BASE_API_URL}/files/{test_file.object_id}'
    dimensions = '100x100'
    extension = 'jpg'

    mock_representations_response = Mock()
    mock_representations_response.json.return_value = {
        'etag': '1',
        'id': test_file.object_id,
        'representations': {
            'entries': [],
        },
        'type': 'file'
    }

    mock_box_session.get.side_effect = [mock_representations_response, mock_content_response]

    thumb = test_file.get_thumbnail_representation(
        dimensions=dimensions,
        extension=extension,
    )

    mock_box_session.get.assert_any_call(
        representation_url,
        headers={'X-Rep-Hints': f'[{extension}?dimensions={dimensions}]'},
        params={'fields': 'representations'},
    )
    assert thumb == b''


def test_get_thumbnail_representation_not_available(
        test_file,
        mock_box_session,
        mock_content_response,
):
    representation_url = f'{API.BASE_API_URL}/files/{test_file.object_id}'
    dimensions = '100x100'
    extension = 'jpg'

    mock_representations_response = Mock()
    mock_representations_response.json.return_value = {
        'etag': '1',
        'id': test_file.object_id,
        'representations': {
            'entries': [
                {
                    'content': {
                        'url_template': 'content_url {+asset_path}'
                    },
                    'info': {
                        'url': 'https://api.box.com/2.0/internal_files/123/versions/345/representations/jpg'
                    },
                    'properties': {},
                    'representation': 'pdf',
                    'status': {'state': 'error', 'code': 'error_password_protected'}
                }
            ]
        },
        'type': 'file'
    }

    mock_box_session.get.side_effect = [mock_representations_response, mock_content_response]

    thumb = test_file.get_thumbnail_representation(
        dimensions=dimensions,
        extension=extension,
    )

    mock_box_session.get.assert_any_call(
        representation_url,
        headers={'X-Rep-Hints': f'[{extension}?dimensions={dimensions}]'},
        params={'fields': 'representations'},
    )
    assert thumb == b''


@pytest.mark.parametrize(
    'disposition_at',
    (
        lazy_fixture('mock_datetime_rfc3339_str'),
        "2035-03-04T10:14:24.000+14:00",
        "2035/03/04 10:14:24.000+14:00",
        lazy_fixture('mock_timezone_aware_datetime_obj'),
    )
)
def test_set_diposition_at(
        test_file,
        mock_box_session,
        disposition_at,
        mock_datetime_rfc3339_str,
):
    expected_url = test_file.get_url()
    expected_data = {'disposition_at': mock_datetime_rfc3339_str}

    test_file.set_disposition_at(disposition_at)

    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_data),
        headers=None,
        params=None,
    )



__all__ = [
    'collaboration',
    'collaboration_allowlist',
    'collaboration_allowlist_entry',
    'collaboration_allowlist_exempt_target',
    'collection',
    'comment',
    'device_pinner',
    'enterprise',
    'events',
    'event',
    'file',
    'file_request',
    'file_version',
    'file_version_retention',
    'folder',
    'folder_lock',
    'group',
    'group_membership',
    'invite',
    'legal_hold',
    'legal_hold_policy',
    'legal_hold_policy_assignment',
    'metadata_cascade_policy',
    'metadata_template',
    'recent_item',
    'retention_policy',
    'retention_policy_assignment',
    'search',
    'sign_request',
    'storage_policy',
    'storage_policy_assignment',
    'terms_of_service',
    'terms_of_service_user_status',
    'trash',
    'task',
    'task_assignment',
    'user',
    'upload_session',
    'webhook',
    'watermark',
    'web_link'
]



from datetime import datetime
from typing import Union, Optional

from dateutil import parser


def normalize_date_to_rfc3339_format(date: Union[None, datetime, str]) -> Optional[str]:
    """
    Normalizes any datetime string supported by dateutil library or a datetime.datetime object to rfc3339 format.

    :param date:  datetime str or datetime object
    :return: date-time str in rfc3339 format
    """
    if date is None:
        return None

    if isinstance(date, str):
        date = parser.parse(date)

    if not isinstance(date, datetime):
        raise TypeError(f"Got unsupported type {date.__class__.__name__!r} for date.")

    timezone_aware_datetime = date if date.tzinfo is not None else date.astimezone()
    return timezone_aware_datetime.isoformat(timespec='seconds')



from test.integration_new import util
from test.integration_new import CLIENT


def test_create_metedata_template_with_fields():
    policy_name = 'Test Retention Policy ' + util.random_name()
    disposition_action = 'permanently_delete'
    retention_length = '1'
    retention_type = 'modifiable'
    policy_description = 'Test Retention Policy'

    retention_policy = CLIENT.create_retention_policy(
        policy_name=policy_name,
        disposition_action=disposition_action,
        retention_length=retention_length,
        retention_type=retention_type,
        description=policy_description
    )

    try:
        assert retention_policy.policy_name == policy_name
        assert retention_policy.disposition_action == disposition_action
        assert retention_policy.retention_length == retention_length
        assert retention_policy.retention_type == retention_type
        assert retention_policy.description == policy_description

        new_policy_name = 'Test Retention Policy ' + util.random_name()
        new_policy_description = 'Test Retention Policy Updated'
        retention_policy.update_info(data={
            'policy_name': new_policy_name,
            'description': new_policy_description
        })

        retention_policy = CLIENT.retention_policy(retention_id=retention_policy.object_id).get()
        assert retention_policy.policy_name == new_policy_name
        assert retention_policy.description == new_policy_description
    finally:
        retention_policy.update_info(data={
            'status': 'retired'
        })



def get_shared_link_header(shared_link: str, password: str = None) -> dict:
    """
    Gets the HTTP header required to use a shared link to grant access to a shared item.

    :param shared_link:
        The shared link.
    :param password:
        The password for the shared link.
    :return:
        The item referred to by the shared link.
    """
    shared_link_password = f'&shared_link_password={password}' if password is not None else ''
    box_api_header = f'shared_link={shared_link}{shared_link_password}'
    return {'BoxApi': box_api_header}



import json
import os
from typing import TYPE_CHECKING, Optional, Iterable, IO

from .base_object import BaseObject
from ..pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..util.api_call_decorator import api_call

if TYPE_CHECKING:
    from boxsdk.object.email_alias import EmailAlias
    from boxsdk.pagination.box_object_collection import BoxObjectCollection
    from boxsdk.object.folder import Folder
    from boxsdk.object.storage_policy_assignment import StoragePolicyAssignment


class User(BaseObject):
    """Represents a Box user."""

    _item_type = 'user'

    @api_call
    def add_email_alias(self, email: str) -> 'EmailAlias':
        """
        Adds a new email alias to the given user's account.

        :param email:
            The email alias to add to the user.
        :returns:
            The new email alias object
        """
        url = self.get_url('email_aliases')
        body = {
            'email': email,
        }
        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_email_aliases(self, limit: Optional[int] = None, fields: Iterable[str] = None) -> 'BoxObjectCollection':
        """
        Gets an list of email aliases for a user.

        :param limit:
            The maximum number of users to return. If not specified, the Box API will determine an appropriate limit.
        :param fields:
            List of fields to request
        :returns:
            An iterator of the user's email aliases
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('email_aliases'),
            limit=limit,
            marker=None,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def remove_email_alias(self, email_alias: 'EmailAlias') -> bool:
        """
        Remove an email alias from the user.

        :param email_alias:
            The email alias to remove.
        :returns:
            Whether the removal succeeded.
        """
        url = self.get_url('email_aliases', email_alias.object_id)
        response = self._session.delete(url, expect_json_response=False)
        return response.ok

    @api_call
    def transfer_content(
            self,
            destination_user: 'User',
            notify: Optional[bool] = None,
            fields: Iterable[str] = None
    ) -> 'Folder':
        """
        Move all of the items owned by a user into a new folder in another user's account.

        :param destination_user:
            The id of the user to transfer content to.
        :param notify:
            Whether the destination user should receive email notification of the transfer.
        :param fields:
            Fields to return on the resulting :class:`Folder` object
        :returns:
            A :class:`Folder` object that was transferred to another user.
        """
        url = self.get_url('folders', '0')
        body = {
            'owned_by': {
                'id': destination_user.object_id,
            },
        }
        params = {}
        if notify is not None:
            params['notify'] = notify
        if fields is not None:
            params['fields'] = ','.join(fields)
        response = self._session.put(url, data=json.dumps(body), params=params).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    def get_storage_policy_assignment(self) -> 'StoragePolicyAssignment':
        """
        Get the storage policy assignment assigned to the user.

        :returns:
            The :class:`StoragePolicyAssignment` object information
        """
        url = self._session.get_url('storage_policy_assignments')
        additional_params = {
            'resolved_for_type': self.object_type,
            'resolved_for_id': self.object_id,
        }
        box_response = self._session.get(url, params=additional_params)
        response = box_response.json()['entries'][0]
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_group_memberships(
            self,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the user group membership using limit-offset paging.

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param offset:
            The offset of the item at which to begin the response.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the groups
        """
        additional_params = {}
        if fields is not None:
            additional_params['fields'] = ','.join(fields)
        return LimitOffsetBasedObjectCollection(
            session=self._session,
            url=self.get_url('memberships'),
            additional_params=additional_params,
            limit=limit,
            offset=offset,
            return_full_pages=False,
        )

    @api_call
    def get_avatar(self) -> bytes:
        """
        Get the avatar for the User.

        :returns:
            Avatar content as bytes.
        """
        url = self.get_url('avatar')
        response = self._session.get(url, expect_json_response=False)
        return response.content

    @api_call
    def upload_avatar(self, image_path: str) -> str:
        """
        Upload avatar image to user account. Supported formats are JPG, JPEG and PNG.
        Maximum allowed file size is 1MB and resolution 1024x1024 pixels.

        :param image_path: Path of the avatar image to upload
        :return: URLs to existing user avatars that were updated
        """
        with open(image_path, 'rb') as image_stream:
            image_extension = os.path.splitext(image_path)[-1][1:]
            return self.upload_avatar_stream(image_stream=image_stream, image_extension=image_extension)

    @api_call
    def upload_avatar_stream(self, image_stream: IO[bytes], image_extension: str) -> str:
        """
        Upload avatar image to user account. Supported formats are JPG, JPEG and PNG.
        Maximum allowed file size is 1MB and resolution 1024x1024 pixels.

        :param image_extension: Extension of the image. Supported: 'jpg', 'jpeg' and 'png'.
        :param image_stream: The file-like object containing the bytes of the image
        :return: URLs to existing user avatars that were updated
        """
        url = self.get_url('avatar')
        files = {'pic': (f'avatar.{image_extension}', image_stream, f'image/{image_extension}')}
        response = self._session.post(url, files=files)
        return response.json()['pic_urls']

    @api_call
    def delete_avatar(self) -> bool:
        """
        Remove avatar from user account

        :return: Whether the deletion succeeded
        """
        url = self.get_url('avatar')
        response = self._session.delete(url)
        return response.ok

    @api_call
    def delete(self, *, notify: bool = True, force: bool = False, **kwargs) -> bool:
        # pylint: disable=arguments-differ,arguments-renamed
        """
        Delete a user's account.  This user will no longer be able to access Box.

        :param notify:
            Whether a notification should be sent about the deletion
        :param force:
            Whether the user should be deleted even if they still own files
        :returns:
            Whether the deletion succeeded
        """
        params = {
            'notify': notify,
            'force': force,
        }
        return super().delete(params=params, **kwargs)



from unittest.mock import Mock
import pytest

from boxsdk.object.legal_hold_policy_assignment import LegalHoldPolicyAssignment
from boxsdk.config import API
from boxsdk.network.default_network import DefaultNetworkResponse


def test_get(test_legal_hold_policy_assignment, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/legal_hold_policy_assignments/{test_legal_hold_policy_assignment.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'legal_hold_policy_assignment',
        'id': test_legal_hold_policy_assignment.object_id,
        'assigned_to': {
            'type': 'user',
            'id': '1234',
        },
    }
    legal_hold_policy_assignment = test_legal_hold_policy_assignment.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(legal_hold_policy_assignment, LegalHoldPolicyAssignment)
    assert legal_hold_policy_assignment.type == 'legal_hold_policy_assignment'
    assert legal_hold_policy_assignment.object_id == test_legal_hold_policy_assignment.object_id
    assert legal_hold_policy_assignment['assigned_to']['type'] == 'user'
    assert legal_hold_policy_assignment['assigned_to']['id'] == '1234'


@pytest.fixture(scope='module')
def delete_assignment_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    return mock_network_response


def test_delete_policy_return_the_correct_response(
        test_legal_hold_policy_assignment,
        mock_box_session,
        delete_assignment_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.delete.return_value = delete_assignment_response
    response = test_legal_hold_policy_assignment.delete()
    # pylint:disable=protected-access
    expected_url = test_legal_hold_policy_assignment.get_url()
    # pylint:enable = protected-access
    mock_box_session.delete.assert_called_once_with(expected_url, params={}, expect_json_response=False, headers=None)
    assert response is True



import json
from typing import Optional, Iterable, TYPE_CHECKING

from boxsdk.util.text_enum import TextEnum
from .base_object import BaseObject
from ..pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection
from ..util.api_call_decorator import api_call
from ..util.default_arg_value import SDK_VALUE_NOT_SET

if TYPE_CHECKING:
    from boxsdk.object.group_membership import GroupMembership
    from boxsdk.object.user import User
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class GroupRole(TextEnum):
    """The role in the group."""
    ADMIN = 'admin'
    MEMBER = 'member'


class Group(BaseObject):
    """Represents a Box group."""

    _item_type = 'group'

    @api_call
    def get_memberships(
            self,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            fields: Optional[Iterable[str]] = None
    ) -> Iterable['GroupMembership']:
        """
        Get the membership records for the group, which indicate which users are included in the group.

        :param limit:
            The maximum number of items to return in a page.
        :param offset:
            The index at which to begin.
        :param fields:
            List of fields to request. If None, will return the default fields for the object.
        :returns:
            The collection of membership objects for the group.
        """
        return LimitOffsetBasedObjectCollection(
            self._session,
            url=self.get_url('memberships'),
            limit=limit,
            offset=offset,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def add_member(
            self,
            user: 'User',
            role: GroupRole = GroupRole.MEMBER,
            configurable_permissions: Optional[str] = SDK_VALUE_NOT_SET
    ) -> 'GroupMembership':
        """
        Add the given user to this group under the given role

        :param user:
            The User to add to the group.
        :param role:
            The role for the user.
        :param configurable_permissions:
            This is a group level permission that is configured for Group members with
            admin role only.
        :returns:
            The new GroupMembership instance.
        """
        url = self._session.get_url('group_memberships')
        body_attributes = {
            'user': {'id': user.object_id},
            'group': {'id': self.object_id},
            'role': role,
        }
        if configurable_permissions is not SDK_VALUE_NOT_SET:
            body_attributes['configurable_permissions'] = configurable_permissions
        box_response = self._session.post(url, data=json.dumps(body_attributes))
        response = box_response.json()
        return self.translator.translate(self._session, response)

    @api_call
    def get_collaborations(
            self,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the collaboration for the group using limit-offset paging.

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param offset:
            The offset of the item at which to begin the response.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the collaboration for the group.
        """
        additional_params = {}
        if fields is not None:
            additional_params['fields'] = ','.join(fields)
        return LimitOffsetBasedObjectCollection(
            session=self._session,
            url=self.get_url('collaborations'),
            additional_params=additional_params,
            limit=limit,
            offset=offset,
            return_full_pages=False,
        )



import sqlalchemy
import uuid
from sqlalchemy.orm import relationship
from test.functional.mock_box.db_model import DbModel


class UserModel(DbModel):
    """DB Model for Box users."""
    __tablename__ = 'box_user'

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, autoincrement=True)  # pylint:disable=invalid-name
    user_id = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: uuid.uuid4().hex)
    name = sqlalchemy.Column(sqlalchemy.String(255))
    login = sqlalchemy.Column(sqlalchemy.String(255))
    owned_files = relationship(
        'FileModel',
        backref='owned_by',
        cascade='save-update, delete',
        foreign_keys='[FileModel.owned_by_id]',
    )
    owned_folders = relationship(
        'FolderModel',
        backref='owned_by',
        cascade='save-update, delete',
        foreign_keys='[FolderModel.owned_by_id]',
    )
    created_collaborations = relationship('CollaborationModel', backref='created_by', cascade='save-update, delete')
    created_locks = relationship('LockModel', backref='created_by', cascade='save-update, delete')
    authorized_application_tokens = relationship('TokenModel', backref='owned_by', cascade='save-update, delete')



# pylint: disable-msg=too-many-locals
import time
from random import randint
from unittest.mock import MagicMock, Mock, call
import io
import json
import pytest

from boxsdk.config import API
from boxsdk.exception import BoxAPIException
from boxsdk.exception import BoxException
from boxsdk.object.file import File
from boxsdk.pagination.limit_offset_based_dict_collection import LimitOffsetBasedDictCollection
from boxsdk.object.upload_session import UploadSession
from boxsdk.util.chunked_uploader import ChunkedUploader


@pytest.fixture()
def test_upload_session(mock_box_session):
    upload_session_response_object = {
        'total_parts': 4,
        'part_size': 2,
        'session_expires_at': '2017-11-09T21:59:16Z',
        'id': 'D5E3F7ADA11A38A0A66AD0B64AACA658',
        'type': 'upload_session',
        'num_parts_processed': 0
    }
    return UploadSession(mock_box_session, 'D5E3F7ADA11A38A0A66AD0B64AACA658', upload_session_response_object)


@pytest.fixture()
def mock_upload_session():
    upload_session_mock_object = Mock(UploadSession)
    upload_session_mock_object.total_parts = 4
    upload_session_mock_object.part_size = 2
    upload_session_mock_object.id = 'F971964745A5CD0C001BZ4E58196BFD'
    upload_session_mock_object.type = 'upload_session'
    upload_session_mock_object.num_parts_processed = 0
    return upload_session_mock_object


def test_start(test_upload_session, mock_box_session):
    expected_put_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}'
    expected_post_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}/commit'
    file_size = 7
    first_sha1 = '2iNhTgJGmg18e9G9q1ycR0sZBNw='
    second_sha1 = 'A0d4GYoEXB7YC+JxzdApt2h09vw='
    third_sha1 = '+CIFFHGVe3u+u4qwiP6b1tFPQmE='
    fourth_sha1 = 'VP0XESCfscB4EJI3QTLGbnniJBs='
    part_one = {
        'part_id': 'CFEB4BA9',
        'offset': 0,
        'size': 2,
        'sha1': first_sha1,
    }
    part_two = {
        'part_id': '4DBB872D',
        'offset': 2,
        'size': 2,
        'sha1': second_sha1,
    }
    part_three = {
        'part_id': '6F2D3486',
        'offset': 4,
        'size': 2,
        'sha1': third_sha1,
    }
    part_four = {
        'part_id': '4DBC872D',
        'offset': 6,
        'size': 1,
        'sha1': fourth_sha1,
    }
    parts = [part_one, part_two, part_three, part_four]
    expected_data = {
        'parts': parts,
    }
    expected_headers = {
        'Content-Type': 'application/json',
        'Digest': 'SHA=L7XhNBn8iSRoZeejJPR27GJOh0A=',
    }

    expected_headers_first_upload = {
        'Content-Type': 'application/octet-stream',
        'Digest': f'SHA={first_sha1}',
        'Content-Range': 'bytes 0-1/7',
    }
    expected_headers_second_upload = {
        'Content-Type': 'application/octet-stream',
        'Digest': f'SHA={second_sha1}',
        'Content-Range': 'bytes 2-3/7',
    }
    expected_headers_third_upload = {
        'Content-Type': 'application/octet-stream',
        'Digest': f'SHA={third_sha1}',
        'Content-Range': 'bytes 4-5/7',
    }
    expected_headers_fourth_upload = {
        'Content-Type': 'application/octet-stream',
        'Digest': f'SHA={fourth_sha1}',
        'Content-Range': 'bytes 6-6/7',
    }

    first_response_mock = Mock()
    second_response_mock = Mock()
    third_response_mock = Mock()
    fourth_response_mock = Mock()

    first_response_mock.json.return_value = {
        'part': part_one
    }
    second_response_mock.json.return_value = {
        'part': part_two
    }
    third_response_mock.json.return_value = {
        'part': part_three
    }
    fourth_response_mock.json.return_value = {
        'part': part_four
    }
    mock_box_session.put.side_effect = [
        first_response_mock,
        second_response_mock,
        third_response_mock,
        fourth_response_mock
    ]
    mock_box_session.post.return_value.json.return_value = {
        'entries': [
            {
                'type': 'file',
                'id': '12345',
                'description': 'This is a test description',
            }
        ]
    }
    flaky_stream = Mock()
    flaky_stream.read.side_effect = [b'ab', None, b'c', b'd', b'ef', b'g', b'']

    chunked_uploader = ChunkedUploader(test_upload_session, flaky_stream, file_size)
    uploaded_file = chunked_uploader.start()
    calls = [
        call(expected_put_url, data=b'ab', headers=expected_headers_first_upload),
        call(expected_put_url, data=b'cd', headers=expected_headers_second_upload),
        call(expected_put_url, data=b'ef', headers=expected_headers_third_upload),
        call(expected_put_url, data=b'g', headers=expected_headers_fourth_upload),
    ]
    flaky_stream.read.assert_has_calls([call(2), call(2), call(2), call(1), call(2), call(2), call(1)], any_order=False)
    mock_box_session.put.assert_has_calls(calls, any_order=True)
    mock_box_session.post.assert_called_once_with(
        expected_post_url,
        data=json.dumps(expected_data),
        headers=expected_headers
    )
    assert uploaded_file.type == 'file'
    assert uploaded_file.id == '12345'
    assert uploaded_file.description == 'This is a test description'
    assert isinstance(uploaded_file, File)
    assert uploaded_file._session == mock_box_session  # pylint:disable=protected-access


def test_resume_cross_process(test_file, mock_upload_session):
    file_size = 7
    part_bytes = b'abcdefg'
    stream = io.BytesIO(part_bytes)
    part_one = {
        'part_id': 'CFEB4BA9',
        'offset': 0,
        'size': 2,
        'sha1': None,
    }
    part_two = {
        'part_id': '4DBB872D',
        'offset': 2,
        'size': 2,
        'sha1': None,
    }
    part_three = {
        'part_id': '6F2D3486',
        'offset': 4,
        'size': 2,
        'sha1': None,
    }
    part_four = {
        'part_id': '4DBC872D',
        'offset': 6,
        'size': 1,
        'sha1': None,
    }
    parts = [part_one, part_two, part_three, part_four]
    mock_upload_session.commit.return_value = test_file
    mock_iterator = MagicMock(LimitOffsetBasedDictCollection)
    mock_iterator.__iter__.return_value = [part_one, part_four]
    mock_upload_session.get_parts.return_value = mock_iterator
    mock_upload_session.upload_part_bytes.side_effect = [part_two, part_three]
    chunked_uploader = ChunkedUploader(mock_upload_session, stream, file_size)
    uploaded_file = chunked_uploader.resume()
    calls = [
        call(offset=2, part_bytes=b'cd', total_size=7),
        call(offset=4, part_bytes=b'ef', total_size=7),
    ]
    mock_upload_session.upload_part_bytes.assert_has_calls(calls, any_order=True)
    mock_upload_session.commit.assert_called_once_with(
        content_sha1=b'/\xb5\xe14\x19\xfc\x89$he\xe7\xa3$\xf4v\xecbN\x87@',
        parts=parts
    )
    assert uploaded_file is test_file


def test_resume_in_process(test_file, mock_upload_session):
    file_size = 7
    part_bytes = b'abcdefg'
    stream = io.BytesIO(part_bytes)
    first_part = {
        'part_id': 'CFEB4BA9',
        'offset': 0,
        'size': 2,
        'sha1': '2iNhTgJGmg18e9G9q1ycR0sZBNw=',
    }
    second_part = {
        'part_id': '4DBB872D',
        'offset': 2,
        'size': 2,
        'sha1': 'A0d4GYoEXB7YC+JxzdApt2h09vw=',
    }
    third_part = {
        'part_id': '6F2D3486',
        'offset': 4,
        'size': 2,
        'sha1': '+CIFFHGVe3u+u4qwiP6b1tFPQmE=',
    }
    fourth_part = {
        'part_id': '4DBC872D',
        'offset': 6,
        'size': 1,
        'sha1': 'VP0XESCfscB4EJI3QTLGbnniJBs=',
    }
    parts = [first_part, second_part, third_part, fourth_part]
    mock_iterator = MagicMock(LimitOffsetBasedDictCollection)
    mock_iterator.__iter__.return_value = [first_part, second_part, fourth_part]
    mock_upload_session.get_parts.return_value = mock_iterator
    mock_upload_session.commit.return_value = test_file
    chunked_uploader = ChunkedUploader(mock_upload_session, stream, file_size)
    mock_upload_session.upload_part_bytes.side_effect = [
        first_part,
        second_part,
        BoxAPIException(502),
        fourth_part
    ]
    uploaded_file = None
    calls = [
        call(offset=0, part_bytes=b'ab', total_size=7),
        call(offset=2, part_bytes=b'cd', total_size=7),
        call(offset=4, part_bytes=b'ef', total_size=7),
        call(offset=6, part_bytes=b'g', total_size=7),
        call(offset=4, part_bytes=b'ef', total_size=7),
    ]
    try:
        chunked_uploader.start()
    except BoxAPIException:
        mock_upload_session.upload_part_bytes.side_effect = [third_part]
        uploaded_file = chunked_uploader.resume()
    mock_upload_session.upload_part_bytes.assert_has_calls(calls, any_order=True)
    mock_upload_session.commit.assert_called_once_with(
        content_sha1=b'/\xb5\xe14\x19\xfc\x89$he\xe7\xa3$\xf4v\xecbN\x87@',
        parts=parts
    )
    assert uploaded_file is test_file


def test_resume_with_upload_random_duration(test_file, mock_upload_session):
    # pylint: disable-msg=unused-argument
    def upload_mock_func(part_bytes, offset, total_size):
        value = upload_results.pop(0)
        time.sleep(randint(0, 1))
        if isinstance(value, Exception):
            raise value
        return value

    file_size = 7
    part_bytes = b'abcdefg'
    stream = io.BytesIO(part_bytes)
    first_part = {
        'part_id': 'CFEB4BA9',
        'offset': 0,
        'size': 2,
        'sha1': '2iNhTgJGmg18e9G9q1ycR0sZBNw=',
    }
    second_part = {
        'part_id': '4DBB872D',
        'offset': 2,
        'size': 2,
        'sha1': 'A0d4GYoEXB7YC+JxzdApt2h09vw=',
    }
    third_part = {
        'part_id': '6F2D3486',
        'offset': 4,
        'size': 2,
        'sha1': '+CIFFHGVe3u+u4qwiP6b1tFPQmE=',
    }
    fourth_part = {
        'part_id': '4DBC872D',
        'offset': 6,
        'size': 1,
        'sha1': 'VP0XESCfscB4EJI3QTLGbnniJBs=',
    }
    parts = [first_part, second_part, third_part, fourth_part]
    mock_iterator = MagicMock(LimitOffsetBasedDictCollection)
    mock_iterator.__iter__.return_value = [first_part, second_part, fourth_part]
    mock_upload_session.get_parts.return_value = mock_iterator
    mock_upload_session.commit.return_value = test_file
    chunked_uploader = ChunkedUploader(mock_upload_session, stream, file_size)
    upload_results = [
        first_part,
        second_part,
        BoxAPIException(502),
        fourth_part,
        third_part
    ]
    mock_upload_session.upload_part_bytes.side_effect = upload_mock_func
    uploaded_file = None
    try:
        chunked_uploader.start()
    except BoxAPIException:
        uploaded_file = chunked_uploader.resume()
    calls = [
        call(offset=0, part_bytes=b'ab', total_size=7),
        call(offset=2, part_bytes=b'cd', total_size=7),
        call(offset=4, part_bytes=b'ef', total_size=7),
        call(offset=6, part_bytes=b'g', total_size=7),
        call(offset=4, part_bytes=b'ef', total_size=7),
    ]
    mock_upload_session.upload_part_bytes.assert_has_calls(calls, any_order=True)
    mock_upload_session.commit.assert_called_once_with(
        content_sha1=b'/\xb5\xe14\x19\xfc\x89$he\xe7\xa3$\xf4v\xecbN\x87@',
        parts=parts
    )
    assert uploaded_file is test_file


def test_abort_with_start(mock_upload_session):
    file_size = 7
    part_bytes = b'abcdefg'
    stream = io.BytesIO(part_bytes)
    chunked_uploader = ChunkedUploader(mock_upload_session, stream, file_size)
    mock_upload_session.abort.return_value = True
    is_aborted = chunked_uploader.abort()
    try:
        chunked_uploader.start()
    except BoxException:
        pass
    mock_upload_session.abort.assert_called_once_with()
    assert is_aborted is True


def test_abort_with_resume(mock_upload_session):
    file_size = 7
    part_bytes = b'abcdefg'
    stream = io.BytesIO(part_bytes)
    chunked_uploader = ChunkedUploader(mock_upload_session, stream, file_size)
    mock_upload_session.abort.return_value = True
    is_aborted = chunked_uploader.abort()
    try:
        chunked_uploader.resume()
    except BoxException:
        pass
    mock_upload_session.abort.assert_called_once_with()
    assert is_aborted is True


def test_resume_after_commit_failed(test_file, mock_upload_session):
    # given
    file_size = 7
    part_bytes = b'abcdefg'
    stream = io.BytesIO(part_bytes)
    first_part = {
        'part_id': 'CFEB4BA9',
        'offset': 0,
        'size': 2,
        'sha1': '2iNhTgJGmg18e9G9q1ycR0sZBNw=',
    }
    second_part = {
        'part_id': '4DBB872D',
        'offset': 2,
        'size': 2,
        'sha1': 'A0d4GYoEXB7YC+JxzdApt2h09vw=',
    }
    third_part = {
        'part_id': '6F2D3486',
        'offset': 4,
        'size': 2,
        'sha1': '+CIFFHGVe3u+u4qwiP6b1tFPQmE=',
    }
    fourth_part = {
        'part_id': '4DBC872D',
        'offset': 6,
        'size': 1,
        'sha1': 'VP0XESCfscB4EJI3QTLGbnniJBs=',
    }
    parts = [first_part, second_part, third_part, fourth_part]
    mock_iterator = MagicMock(LimitOffsetBasedDictCollection)
    mock_iterator.__iter__.return_value = [first_part, second_part, third_part, fourth_part]
    mock_upload_session.get_parts.return_value = mock_iterator
    mock_upload_session.commit.side_effect = [None, test_file]
    chunked_uploader = ChunkedUploader(mock_upload_session, stream, file_size)
    mock_upload_session.upload_part_bytes.side_effect = [
        first_part,
        second_part,
        third_part,
        fourth_part
    ]

    # when
    uploaded_file = chunked_uploader.start()

    # then
    assert uploaded_file is None

    # when
    uploaded_file = chunked_uploader.resume()

    # then
    expected_call = call(
        content_sha1=b'/\xb5\xe14\x19\xfc\x89$he\xe7\xa3$\xf4v\xecbN\x87@',
        parts=parts
    )
    mock_upload_session.commit.assert_has_calls([expected_call, expected_call])
    assert uploaded_file is test_file



from unittest.mock import Mock
import json
import pytest

from boxsdk.config import API
from boxsdk.object.task_assignment import TaskAssignment, ResolutionState
from boxsdk.network.default_network import DefaultNetworkResponse


@pytest.fixture(scope='module')
def delete_task_assignment_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    return mock_network_response


def test_get_assignment(test_task_assignment, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/task_assignments/{test_task_assignment.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': test_task_assignment.object_type,
        'id': test_task_assignment.object_id,
        'assigned_to': {
            'type': 'user',
            'id': '11111',
        },
    }
    retrieved_task = test_task_assignment.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(retrieved_task, TaskAssignment)
    assert retrieved_task.object_type == test_task_assignment.object_type
    assert retrieved_task.object_id == test_task_assignment.object_id
    assert retrieved_task.assigned_to['type'] == 'user'
    assert retrieved_task.assigned_to['id'] == '11111'


def test_delete_policy_return_the_correct_response(
        test_task_assignment,
        mock_box_session,
        delete_task_assignment_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.delete.return_value = delete_task_assignment_response
    response = test_task_assignment.delete()
    # pylint:disable=protected-access
    expected_url = f'{API.BASE_API_URL}/task_assignments/{test_task_assignment.object_id}'
    # pylint:enable = protected-access
    mock_box_session.delete.assert_called_once_with(expected_url, params={}, expect_json_response=False, headers=None)
    assert response is True


def test_update(test_task_assignment, mock_box_session):
    new_message = 'New Message'
    resolution_state = ResolutionState.APPROVED
    expected_url = f'{API.BASE_API_URL}/task_assignments/{test_task_assignment.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': 'task_assignment',
        'id': test_task_assignment.object_id,
        'message': new_message,
        'resolution_state': resolution_state,
    }
    expected_body = {
        'message': new_message,
        'resolution_state': resolution_state,
    }
    updated_task_assignment = test_task_assignment.update_info(data={'message': new_message, 'resolution_state': resolution_state})
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(expected_body), headers=None, params=None)
    assert isinstance(updated_task_assignment, TaskAssignment)
    assert updated_task_assignment.message == new_message
    assert updated_task_assignment.object_type == test_task_assignment.object_type
    assert updated_task_assignment.object_id == test_task_assignment.object_id
    assert updated_task_assignment.resolution_state == resolution_state



import os
from typing import Any

from test.integration_new import util


class LocalLargeFile:

    def __init__(self, *, name: str = None):
        if name is None:
            name = util.random_name()
        with open(name, 'wb') as f:
            f.seek(1024 * 1024 * 21)
            f.write(b'0')
        self.path = name

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if os.path.exists(self.path):
            os.remove(self.path)
